"""
Parameter constants
===================

Defines the strings of different json parameters
Using variables rather than strings is helpful for typo prevention. A typo in a variable will
raise an error where the variable was misspelled, whereas a typo in a string (dictionnary key)
do not nessarily raise an error immediately, making the typo harder to fix.
"""

#####################
# 1st level of json #
#####################

ECONOMIC_DATA = "economic_data"
PROJECT_DATA = "project_data"
SIMULATION_SETTINGS = "simulation_settings"
FIX_COST = "fixcost"

# Asset groups
ENERGY_CONVERSION = "energyConversion"
ENERGY_CONSUMPTION = "energyConsumption"
ENERGY_PRODUCTION = "energyProduction"
ENERGY_STORAGE = "energyStorage"
ENERGY_BUSSES = "energyBusses"
ENERGY_PROVIDERS = "energyProviders"

##### For D1 ######################
# Definition of allowed oemof types
OEMOF_TRANSFORMER = "transformer"
OEMOF_GEN_STORAGE = "storage"
OEMOF_SOURCE = "source"
OEMOF_SINK = "sink"
OEMOF_BUSSES = "bus"
OEMOF_ExtractionTurbineCHP = "extractionTurbineCHP"

# Dict generated from above defined strings
ACCEPTED_ASSETS_FOR_ASSET_GROUPS = {
    ENERGY_CONVERSION: [OEMOF_TRANSFORMER, OEMOF_ExtractionTurbineCHP],
    ENERGY_STORAGE: [OEMOF_GEN_STORAGE],
    ENERGY_PRODUCTION: [OEMOF_SOURCE],
    ENERGY_CONSUMPTION: [OEMOF_SINK],
}

##############################################
# Constant variable definitions (csv + json) #
##############################################

# Central constant variables
UNIT = "unit"
VALUE = "value"
DATA = "data"

# Economic parameters
CURR = "currency"
DISCOUNTFACTOR = "discount_factor"
LABEL = "label"
PROJECT_DURATION = "project_duration"
TAX = "tax"
FILENAME = "file_name"
STORAGE_FILENAME = "storage_filename"

# Simulation settings: Place and time
EVALUATED_PERIOD = "evaluated_period"
START_DATE = "start_date"
TIMESTEP = "timestep"
PERIODS = "periods"
LONGITUDE = "longitude"
LATITUDE = "latitude"

# Project data and simulation settings (true/false)
OUTPUT_LP_FILE = "output_lp_file"
PROJECT_NAME = "project_name"
SCENARIO_NAME = "scenario_name"
COUNTRY = "country"
PROJECT_ID = "project_id"
SCENARIO_ID = "scenario_id"
SCENARIO_DESCRIPTION = "scenario_description"

# Asset definitions
DSM = "dsm"
TYPE_ASSET = "type_asset"
EFFICIENCY = "efficiency"
OPTIMIZE_CAP = "optimizeCap"
INSTALLED_CAP = "installedCap"
MAXIMUM_CAP = "maximumCap"
AGE_INSTALLED = "age_installed"
LIFETIME = "lifetime"
DEVELOPMENT_COSTS = "development_costs"
SPECIFIC_COSTS = "specific_costs"
SPECIFIC_COSTS_OM = "specific_costs_om"
DISPATCH_PRICE = "dispatch_price"
OEMOF_ASSET_TYPE = "type_oemof"
BETA = "beta"

# Specific parameters
RENEWABLE_ASSET_BOOL = "renewableAsset"
RENEWABLE_SHARE_DSO = "renewable_share"
EMISSION_FACTOR = "emission_factor"

# Asset definitions: Providers
ENERGY_PRICE = "energy_price"
FEEDIN_TARIFF = "feedin_tariff"
PEAK_DEMAND_PRICING = "peak_demand_pricing"
PEAK_DEMAND_PRICING_PERIOD = "peak_demand_pricing_period"

# Asset definitions: Storage
C_RATE = "c_rate"
INPUT_POWER = "input power"
OUTPUT_POWER = "output power"
STORAGE_CAPACITY = "storage capacity"
SOC_INITIAL = "soc_initial"
SOC_MAX = "soc_max"
SOC_MIN = "soc_min"
THERM_LOSSES_REL = "fixed_thermal_losses_relative"
THERM_LOSSES_ABS = "fixed_thermal_losses_absolute"

# Constraints
CONSTRAINTS = "constraints"
MINIMAL_RENEWABLE_FACTOR = "minimal_renewable_factor"
MAXIMUM_EMISSIONS = "maximum_emissions"
MINIMAL_DEGREE_OF_AUTONOMY = "minimal_degree_of_autonomy"
NET_ZERO_ENERGY = "net_zero_energy"

#######################################
# Parameters added in pre-processing #
#######################################
# Units
UNIT_YEAR = "year"
UNIT_HOUR = "hour"
UNIT_MINUTE = "min"
UNIT_EMISSIONS = "kgCO2eq/a"
UNIT_SPECIFIC_EMISSIONS = "kgCO2eq/kWheleq"

# Preprocessing: Time
END_DATE = "end_date"
TIME_INDEX = "time_index"
TIMESERIES = "timeseries"
TIMESERIES_NORMALIZED = "timeseries_normalized"
TIMESERIES_PEAK = "timeseries_peak"
TIMESERIES_TOTAL = "timeseries_total"
TIMESERIES_AVERAGE = "timeseries_average"

# Other pre-processing
VERSION_NUM = "Version number"

# Pre-processing cost parameters
ANNUITY_FACTOR = "annuity_factor"
CRF = "CRF"

# Processed cost parameters
LIFETIME_SPECIFIC_COST_OM = "lifetime_specific_cost_om"
LIFETIME_PRICE_DISPATCH = "lifetime_price_dispatch"
LIFETIME_SPECIFIC_COST = "lifetime_specific_cost"
ANNUITY_SPECIFIC_INVESTMENT_AND_OM = (
    "annuity_of_specific_investment_costs_and_specific_annual_om"
)
SIMULATION_ANNUITY = "simulation_annuity"

SPECIFIC_REPLACEMENT_COSTS_INSTALLED = (
    "specific_replacement_costs_of_installed_capacity"
)

SPECIFIC_REPLACEMENT_COSTS_OPTIMIZED = (
    "specific_replacement_costs_of_optimized_capacity"
)

# Other Parameters
LES_ENERGY_VECTOR_S = "energy_vectors_of_the_local_energy_system"
OUTFLOW_DIRECTION = "outflow_direction"
INFLOW_DIRECTION = "inflow_direction"
ENERGY_VECTOR = "energyVector"
EXCESS_SINK = "_excess"
# The maximum additional capacity of an asset (MAXIMUM_ADD_CAP = MAXIMUM_CAP - INSTALLED_CAP)
MAXIMUM_ADD_CAP = "maximumAddCap"
MAXIMUM_CAP_NORMALIZED = "maximumCap_normalized"
MAXIMUM_ADD_CAP_NORMALIZED = "maximumAddCap_normalized"
INSTALLED_CAP_NORMALIZED = "installedCap_normalized"

# DSO
DSO_CONSUMPTION = "_consumption"
DSO_FEEDIN = "_feedin"
DSO_PEAK_DEMAND_SUFFIX = "_pdp"  # short for peak demand pricing
DSO_PEAK_DEMAND_PERIOD = "_period"
CONNECTED_CONSUMPTION_SOURCE = "connected_consumption_sources"
CONNECTED_PEAK_DEMAND_PRICING_TRANSFORMERS = (
    "connected_peak_demand_pricing_transformers"
)
CONNECTED_FEEDIN_SINK = "connected_feedin_sink"

# Autogenerated assets
DISPATCHABILITY = "dispatchable"
AVAILABILITY_DISPATCH = "availability_timeseries"
ASSET_DICT = "asset_list"
#######################################
# Parameters added in post-processing #
#######################################

SIMULATION_RESULTS = "simulation_results"

# oemof simulation parameters:
OBJECTIVE_VALUE = "objective_value"
SIMULTATION_TIME = "simulation_time"
MODELLING_TIME = "modelling_time"

# Logs
LOGS = "logs"
ERRORS = "errors"
WARNINGS = "warnings"

# Names for KPI output
KPI = "kpi"
KPI_SCALARS_DICT = "scalars"
KPI_UNCOUPLED_DICT = "KPI_individual_sectors"
KPI_COST_MATRIX = "cost_matrix"
KPI_SCALAR_MATRIX = "scalar_matrix"

# Flows
FLOW = "flow"
OPTIMIZED_FLOWS = "optimizedFlows"
TOTAL_FLOW = "total_flow"
ANNUAL_TOTAL_FLOW = "annual_total_flow"
PEAK_FLOW = "peak_flow"
AVERAGE_FLOW = "average_flow"
TIMESERIES_SOC = "timeseries_soc"
AVERAGE_SOC = "average_soc"

# Capacity
OPTIMIZED_ADD_CAP = "optimizedAddCap"

# Costs - Annuities
ANNUITY_OM = "annuity_om"
ANNUITY_TOTAL = "annuity_total"

# Costs - Total per asset
COST_TOTAL = "costs_total"
COST_OPERATIONAL_TOTAL = "costs_om_total"
COST_OM = "costs_cost_om"  # Fix asset operation/management costs per year, not depending on use
COST_DISPATCH = (
    "costs_dispatch"  # Variable asset operation/management costs, depending on dispatch
)
COST_INVESTMENT = "costs_investment_over_lifetime"
COST_UPFRONT = "costs_upfront_in_year_zero"
COST_REPLACEMENT = "replacement_costs_during_project_lifetime"

# Levelized cost of electricity
LCOE_ASSET = "levelized_cost_of_energy_of_asset"

# Other KPI
TOTAL_RENEWABLE_GENERATION_IN_LES = "total_internal_renewable_generation"
TOTAL_NON_RENEWABLE_GENERATION_IN_LES = "total_internal_non-renewable_generation"
TOTAL_GENERATION_IN_LES = "total_internal_generation"
TOTAL_RENEWABLE_ENERGY_USE = "total_renewable_energy_use"
TOTAL_NON_RENEWABLE_ENERGY_USE = "total_non-renewable_energy_use"
RENEWABLE_FACTOR = "renewable_factor"
RENEWABLE_SHARE_OF_LOCAL_GENERATION = "renewable_share_of_local_generation"
TOTAL_EMISSIONS = "total_emissions"
SPECIFIC_EMISSIONS_ELEQ = "specific_emissions_per_electricity_equivalent"

TOTAL_DEMAND = "total_demand"
TOTAL_EXCESS = "total_excess"
TOTAL_FEEDIN = "total_feedin"
TOTAL_CONSUMPTION_FROM_PROVIDERS = "total_consumption_from_energy_provider"
SUFFIX_ELECTRICITY_EQUIVALENT = "_electricity_equivalent"
ATTRIBUTED_COSTS = "attributed_costs"
LCOeleq = "levelized_costs_of_electricity_equivalent"

DEGREE_OF_SECTOR_COUPLING = "degree_of_sector_coupling"
DEGREE_OF_AUTONOMY = "degree_of_autonomy"
ONSITE_ENERGY_FRACTION = "onsite_energy_fraction"
ONSITE_ENERGY_MATCHING = "onsite_energy_matching"
DEGREE_OF_NZE = "degree_of_NZE"

# KPI_FLOW_MATRIX
KPI_SCALARS = (
    ANNUITY_OM,
    ANNUITY_TOTAL,
    COST_INVESTMENT,
    COST_OPERATIONAL_TOTAL,
    COST_OM,
    COST_DISPATCH,
    COST_TOTAL,
    COST_UPFRONT,
)

DEMANDS = "demands"
RESOURCES = "resources"



"""
Module E2 - Economic processing
===============================

The module processes the simulation results regarding economic parameters:
- calculate lifetime expenditures based on variable energy flows
- calculate lifetime investment costs
- calculate present value of an asset
- calculate revenue
- calculate yearly cash flows of whole project for project lifetime (cash flow projection)
- calculate fuel price expenditures calculate upfront investment costs
- calculate operation management costs (FOM)
- calculate upfront investment costs (UIC)
- calculate annuity per asset
- calculate annuity for the whole project
- calculate net present value
- calculate levelised cost of energy
- calculate levelised cost of energy carriers (electricity, H2, heat)
"""

import logging
import pandas as pd
import warnings

from multi_vector_simulator.utils.constants_json_strings import (
    UNIT,
    CURR,
    UNIT_YEAR,
    VALUE,
    ECONOMIC_DATA,
    CURR,
    LABEL,
    DEVELOPMENT_COSTS,
    SPECIFIC_COSTS,
    INSTALLED_CAP,
    SIMULATION_SETTINGS,
    OUTFLOW_DIRECTION,
    LIFETIME_SPECIFIC_COST,
    CRF,
    LIFETIME_SPECIFIC_COST_OM,
    LIFETIME_PRICE_DISPATCH,
    ANNUAL_TOTAL_FLOW,
    OPTIMIZED_ADD_CAP,
    OUTFLOW_DIRECTION,
    ANNUITY_OM,
    ANNUITY_TOTAL,
    COST_TOTAL,
    COST_OPERATIONAL_TOTAL,
    COST_INVESTMENT,
    COST_DISPATCH,
    COST_OM,
    COST_UPFRONT,
    ENERGY_PRODUCTION,
    TOTAL_FLOW,
    ENERGY_CONSUMPTION,
    ENERGY_CONVERSION,
    ENERGY_STORAGE,
    LCOE_ASSET,
    OUTPUT_POWER,
    INPUT_POWER,
    STORAGE_CAPACITY,
    COST_REPLACEMENT,
    SIMULATION_RESULTS,
    FLOW,
    SPECIFIC_REPLACEMENT_COSTS_INSTALLED,
    SPECIFIC_REPLACEMENT_COSTS_OPTIMIZED,
)


class MissingParametersForEconomicEvaluation(UserWarning):
    "Warning if one or more parameters are missing for economic post-processing  of an asset"
    pass


def get_costs(dict_asset, economic_data):
    r"""
    Calculates economic KPI of the asset handed to the function

    Parameters
    ----------
    dict_asset: dict
        Asset to be evaluated.
        Warning messages in place in case that the asset should not be evaluated.

    economic_data: dict
        Economic data of the project

    Returns
    -------
    Updated dict_asset with following KPI:
    - COST_INVESTMENT
    - COST_UPFRONT
    - COST_REPLACEMENT
    - COST_TOTAL
    - COST_OM
    - COST_DISPATCH
    - COST_OPERATIONAL_TOTAL
    - ANNUITY_TOTAL
    - ANNUITY_OM

    Tested with:
    - test_all_cost_info_parameters_added_to_dict_asset()
    - Test_Economic_KPI.test_benchmark_Economic_KPI_C2_E2()

    """

    logging.debug("Calculating costs of asset %s", dict_asset[LABEL])

    # helper for developing get_costs() and E modules
    if not isinstance(dict_asset, dict):
        logging.warning(
            f"Function E2.get_costs() is used on {dict_asset}, eventhough it is not a dict. Check loops in E modules."
        )

    # helper for developing get_costs() and E modules
    if dict_asset[LABEL] in [ECONOMIC_DATA, SIMULATION_SETTINGS, SIMULATION_RESULTS]:
        logging.warning(
            f"Function E2.get_costs() is used on {dict_asset[LABEL]}, eventhough it should not be applied to it. Check loops in E modules."
        )

    # Testing, if the dict_asset includes all parameters necessary for the proceeding evaluation
    all_list_in_dict(
        dict_asset,
        [
            LIFETIME_SPECIFIC_COST,
            OPTIMIZED_ADD_CAP,
            DEVELOPMENT_COSTS,
            SPECIFIC_COSTS,
            LIFETIME_PRICE_DISPATCH,
            FLOW,
        ],
    )

    # Part of the investment costs to be paid upfront at t=0
    costs_investment_upfront = calculate_costs_upfront_investment(
        capacity=dict_asset[OPTIMIZED_ADD_CAP][VALUE],
        specific_cost=dict_asset[SPECIFIC_COSTS][VALUE],
        development_costs=dict_asset[DEVELOPMENT_COSTS][VALUE],
    )

    dict_asset.update(
        {COST_UPFRONT: {VALUE: costs_investment_upfront, UNIT: economic_data[CURR]}}
    )

    # Part of the investment costs to be paid due to replacements
    costs_replacement = calculate_costs_replacement(
        specific_replacement_of_initial_capacity=dict_asset[
            SPECIFIC_REPLACEMENT_COSTS_INSTALLED
        ][VALUE],
        specific_replacement_of_optimized_capacity=dict_asset[
            SPECIFIC_REPLACEMENT_COSTS_OPTIMIZED
        ][VALUE],
        initial_capacity=dict_asset[INSTALLED_CAP][VALUE],
        optimized_capacity=dict_asset[OPTIMIZED_ADD_CAP][VALUE],
    )

    dict_asset.update(
        {COST_REPLACEMENT: {VALUE: costs_replacement, UNIT: economic_data[CURR]}}
    )

    # Total investment costs including investments into the asset, replacement costs and development costs
    costs_investment_lifetime = calculate_total_capital_costs(
        upfront=dict_asset[COST_UPFRONT][VALUE],
        replacement=dict_asset[COST_REPLACEMENT][VALUE],
    )

    dict_asset.update(
        {COST_INVESTMENT: {VALUE: costs_investment_lifetime, UNIT: economic_data[CURR]}}
    )

    # Operation and management expenditures over the project lifetime
    operation_and_management_expenditures = calculate_operation_and_management_expenditures(
        specific_om_cost=dict_asset[LIFETIME_SPECIFIC_COST_OM][VALUE],
        installed_capacity=dict_asset[INSTALLED_CAP][VALUE],
        optimized_add_capacity=dict_asset[OPTIMIZED_ADD_CAP][VALUE],
    )
    dict_asset.update(
        {
            COST_OM: {
                VALUE: operation_and_management_expenditures,
                UNIT: economic_data[CURR],
            }
        }
    )

    # Dispatch expenditures of the asset over the project lifetime
    if isinstance(dict_asset.get(OUTFLOW_DIRECTION, None), list):
        costs_dispatch = 0
        for bus in dict_asset[OUTFLOW_DIRECTION]:
            costs_dispatch += calculate_dispatch_expenditures(
                dispatch_price=dict_asset[LIFETIME_PRICE_DISPATCH][VALUE],
                flow=dict_asset[FLOW][bus],
                asset=dict_asset[LABEL],
            )
    else:
        costs_dispatch = calculate_dispatch_expenditures(
            dispatch_price=dict_asset[LIFETIME_PRICE_DISPATCH][VALUE],
            flow=dict_asset[FLOW],
            asset=dict_asset[LABEL],
        )

    dict_asset.update(
        {COST_DISPATCH: {VALUE: costs_dispatch, UNIT: economic_data[CURR]}}
    )

    # Total operational expenditures over the lifetime
    total_operational_expenditures = calculate_total_operational_expenditures(
        dict_asset[COST_OM][VALUE], dict_asset[COST_DISPATCH][VALUE]
    )
    dict_asset.update({COST_OPERATIONAL_TOTAL: {VALUE: total_operational_expenditures}})

    # Total costs of the assets, capital and operational
    total_asset_costs_over_lifetime = calculate_total_asset_costs_over_lifetime(
        dict_asset[COST_INVESTMENT][VALUE], dict_asset[COST_OPERATIONAL_TOTAL][VALUE]
    )
    dict_asset.update(
        {
            COST_TOTAL: {
                VALUE: total_asset_costs_over_lifetime,
                UNIT: economic_data[CURR],
            }
        }
    )

    dict_asset.update(
        {
            ANNUITY_TOTAL: {
                VALUE: dict_asset[COST_TOTAL][VALUE] * economic_data[CRF][VALUE],
                UNIT: CURR + "/" + UNIT_YEAR,
            },
            ANNUITY_OM: {
                VALUE: dict_asset[COST_OPERATIONAL_TOTAL][VALUE]
                * economic_data[CRF][VALUE],
                UNIT: CURR + "/" + UNIT_YEAR,
            },
        }
    )


def calculate_total_asset_costs_over_lifetime(
    costs_investment, cost_operational_expenditures
):
    r"""
    Calculate costs of an asset over whole lifetime

    Parameters
    ----------
    costs_investment: float
        Investment costs over whole lifetime

    cost_operational_expenditures: float
        Operation and management as well as dispatch expenditures over whole lifetime

    Returns
    -------
    total_asset_costs_over_lifetime: float
        costs of an asset over whole lifetime, including upfront investment costs, development costs, replacement costs, operation and management expenditures, dispatch expenditures
    """
    total_asset_costs_over_lifetime = costs_investment + cost_operational_expenditures
    return total_asset_costs_over_lifetime


def calculate_dispatch_expenditures(dispatch_price, flow, asset):
    r"""
    Calculate the expenditures connected to an asset due to its dispatch

    Parameters
    ----------
    dispatch_price: float, int or pd.Series
        Dispatch price of an asset (variable costs), ie. how much has to be paid for each unit of dispatch
        Raises error if type does not match
        a) lifetime_price_dispatch (taking into account all years of operation)
        b) price_dispatch (taking into account one year of operation)

    flow: pd.Series
        Dispatch of the asset

    asset: str
        Label of the asset

    Returns
    -------
    a) Total dispatch expenditures of an asset
    b) Annual dispatch expenditures of an asset
    """

    if isinstance(dispatch_price, float) or isinstance(dispatch_price, int):
        # Dispatch price is a scalar
        dispatch_expenditures = dispatch_price * sum(flow)
    elif isinstance(dispatch_price, pd.Series):
        # Dispatch price is defined as a timeseries
        dispatch_expenditures = sum(dispatch_price * flow)
    elif isinstance(dispatch_price, list):
        # Dispatch price is defined as a list, ie. the asset has multiple flows
        dispatch_expenditures = 0
        for list_price in dispatch_price:
            partial_dispatch_expenditures = calculate_dispatch_expenditures(
                list_price, flow, asset + " (list entry)"
            )
            dispatch_expenditures += partial_dispatch_expenditures
    else:
        raise TypeError(
            f"The dispatch price of asset {asset} is neither float, list nor pd.Series but {type(dispatch_price)}."
            f"Please adapt E2.calculate_dispatch_costs() to evaluate the dispatch_expenditures of the asset."
        )

    return dispatch_expenditures


def calculate_costs_upfront_investment(specific_cost, capacity, development_costs):
    r"""
    Calculate investment costs of an asset
    Depending on the specific_cost provided,
    either the total asset's lifetime investment costs or the upfront investment costs are calculated,

    Parameters
    ----------
    specific_cost: float
        a) Specific per-unit investment costs of an asset over its lifetime, including all replacement costs
        b) Specific per-unit investment costs of an asset in year 0

    capacity: float
        Capacity to be installed

    development_costs: float
        Fix development costs, ie. an expense not related to the capacity that is installed. Could be planning costs of the asset.

    Returns
    -------
    costs_investment: float
        a) Total investment costs of an asset over its lifetime, including all replacement costs
        b) Upfront investment costs in year 0

    """
    costs_upfront_investment = specific_cost * capacity + development_costs
    return costs_upfront_investment


def calculate_total_capital_costs(upfront, replacement):
    r"""
    Calculate total capital expenditures

    Parameters
    ----------
    upfront: float
        Upfront investments at t=0, including development costs

    replacement: float
        Replacement costs of pre-installed and new assets


    Returns
    -------
    cost_total_investment: float
        Total capital costs
    """
    cost_total_investment = upfront + replacement
    return cost_total_investment


def calculate_costs_replacement(
    specific_replacement_of_initial_capacity,
    specific_replacement_of_optimized_capacity,
    initial_capacity,
    optimized_capacity,
):
    r"""
    Calculate (the present value of) the replacement costs over the project lifetime

    Parameters
    ----------
    specific_replacement_of_initial_capacity: float
        Per-unit replacement costs of an asset that was pre-existing at the location

    specific_replacement_of_optimized_capacity: float
        Per-unit replacement costs of an asset that is to be installed

    initial_capacity: float
        Initial capacity installed

    optimized_capacity: float
        Additional capacity to be installed, as optimized

    Returns
    -------
    costs_replacements: float
        Aggregated replacement costs over the project lifetime
    """
    costs_replacements = (
        specific_replacement_of_initial_capacity * initial_capacity
        + specific_replacement_of_optimized_capacity * optimized_capacity
    )
    return costs_replacements


def calculate_operation_and_management_expenditures(
    specific_om_cost, installed_capacity, optimized_add_capacity
):
    r"""
    Calculate operation and management expenditures

    Parameters
    ----------
    specific_om_cost: float
        a) specific operation and management costs per unit in year 0
        b) specific operation and management costs per unit for the whole project lifetime

    installed_capacity: float
        Capacity installed initially

    optimized_add_capacity: float
        Capacity installed within the optimization scenario

    Returns
    -------
    costs_operation_and_management: float
        a) Operation and management expenditures in year 0
        b) Total operation and management expenditures over the project lifetime
    """
    costs_operation_and_management = specific_om_cost * (
        installed_capacity + optimized_add_capacity
    )
    return costs_operation_and_management


def calculate_total_operational_expenditures(
    operation_and_management_expenditures, dispatch_expenditures
):
    r"""
    Calculate total expenditures of an asset (operational costs)

    Parameters
    ----------
    operation_and_management_expenditures: float
        a) operation and management expenditures per annum for the installed capacity
        b) operation and management expenditures for whole project lifetime for the installed capacity

    dispatch_expenditures: float
        a) dispatch expenditures per annum for the installed capacity
        b) dispatch expenditures for whole project lifetime for the installed capacity

    Returns
    -------
    total_operational_expenditures: float
        a) total operational expenditures per annum for installed capacity
        b) total operational expenditures for whole project lifetime for installed capacity
    """
    total_operational_expenditures = (
        operation_and_management_expenditures + dispatch_expenditures
    )
    return total_operational_expenditures


def all_list_in_dict(dict_asset, list):
    r"""
    Checks if all items of a list are withing the keys of a dictionary

    Parameters
    ----------
    dict_asset: dict
        Dict with the keys to be evaluated

    list: list
        List of keys (parameter in strings) that should be in dict

    Returns
    -------
    boolean: bool
        True: All items in keys of the dict
        False: At least one item is not in keys of the dict
    """
    boolean = all([name in dict_asset for name in list]) is True
    if boolean is False:
        missing_parameters = []
        for name in list:
            if name not in dict_asset:
                missing_parameters.append(name)
        missing_parameters = ", ".join(map(str, missing_parameters))
        raise MissingParametersForEconomicEvaluation(
            f"Asset {dict_asset[LABEL]} is missing parameters for the economic evaluation: {missing_parameters}."
            f"These parameters are needed for E2.get_costs(). Please check the E modules."
        )
    return boolean


def lcoe_assets(dict_asset, asset_group):
    r"""
    Calculates the levelized cost of electricity (lcoe) of each asset. [Follow this link for information](docs/MVS_Outputs.rst)

    Parameters
    ----------
    dict_asset: dict
        Dictionary defining an asset

    asset_group: str
        Defining to which asset group the asset belongs

    Returns
    -------
    Updates the asset dictionary with the calculated LCOE_ASSET.
    Storages have four values LCOE_ASSET: One for the overall storage including all costs, and one each for the components.

    Notes
    -----

    .. math::

        LCOE\_ASSET = \frac{A}{ E_{throughput} } \\
        \textrm{If } E_{throughput} = 0, LCOE\_ASSET = 0

    """

    lcoe_a = 0

    if asset_group == ENERGY_STORAGE:
        if dict_asset[OUTPUT_POWER][TOTAL_FLOW][VALUE] > 0:
            storage_annuity = (
                dict_asset[INPUT_POWER][ANNUITY_TOTAL][VALUE]
                + dict_asset[OUTPUT_POWER][ANNUITY_TOTAL][VALUE]
                + dict_asset[STORAGE_CAPACITY][ANNUITY_TOTAL][VALUE]
            )
            lcoe_a = storage_annuity / dict_asset[OUTPUT_POWER][TOTAL_FLOW][VALUE]

        for component in [INPUT_POWER, OUTPUT_POWER]:
            if dict_asset[component][TOTAL_FLOW][VALUE] > 0:
                lcoe_a_component = (
                    dict_asset[component][ANNUITY_TOTAL][VALUE]
                    / dict_asset[component][TOTAL_FLOW][VALUE]
                )
                dict_asset[component].update(
                    {LCOE_ASSET: {VALUE: lcoe_a_component, UNIT: CURR + "/kWh"}}
                )
                if component == INPUT_POWER:
                    lcoe_a_component = (
                        dict_asset[STORAGE_CAPACITY][ANNUITY_TOTAL][VALUE]
                        / dict_asset[component][TOTAL_FLOW][VALUE]
                    )
                    dict_asset[STORAGE_CAPACITY].update(
                        {LCOE_ASSET: {VALUE: lcoe_a_component, UNIT: CURR + "/kWh"}}
                    )
            else:
                dict_asset[component].update(
                    {LCOE_ASSET: {VALUE: 0, UNIT: CURR + "/kWh"}}
                )
                if component == INPUT_POWER:
                    dict_asset[STORAGE_CAPACITY].update(
                        {LCOE_ASSET: {VALUE: 0, UNIT: CURR + "/kWh"}}
                    )

    elif dict_asset[TOTAL_FLOW][VALUE] > 0:
        lcoe_a = dict_asset[ANNUITY_TOTAL][VALUE] / dict_asset[TOTAL_FLOW][VALUE]

    dict_asset.update({LCOE_ASSET: {VALUE: lcoe_a, UNIT: CURR + "/kWh"}})



# mvs_tool -i tests/benchmark_test_inputs/Feature_input_flows_as_list -ext csv -f
# mvs_tool -i tests/benchmark_test_inputs/Feature_output_flows_as_list -ext csv -f
import oemof.solph as solph
from oemof.solph import Flow

FLOW_DEFAULT = {
    "values": None,
    "max": [],
    "variable_costs": "0",  # []
    "nonconvex": None,
    "summed_min": None,
    "nominal_value": None,
    "investment": None,
    "fix": "None",  # []
    "negative_gradient": {"ub": [], "costs": 0},
    "summed_max": None,
    "integer": None,
    "positive_gradient": {"ub": [], "costs": 0},
    "min": [],
}
flow_default = FLOW_DEFAULT

INVESTMENT_DEFAULT = solph.options.Investment().__dict__


def parse_investment(inv):
    investment_arguments = []
    for p, pval in inv.__dict__.items():
        if p in INVESTMENT_DEFAULT:
            if INVESTMENT_DEFAULT[p] != pval:
                investment_arguments.append(f"{p}={pval}")
        else:
            investment_arguments.append(f"{p}={pval}")
    return f"solph.options.Investment({', '.join(investment_arguments)})"


def is_bus_flow_dict(d):
    answer = False
    if isinstance(d, dict):
        answer = 0
        for k, v in d.items():
            if isinstance(k, solph.Bus) and isinstance(v, solph.Flow):
                pass
            else:
                answer = answer + 1
        answer = answer == 0
    return answer


def parse_bus_flow_dict(d):
    bus_flow_pairs = []
    for bus, flow in d.items():
        flow_arguments = []
        for p, pval in flow.__dict__.items():

            if p in FLOW_DEFAULT:
                # print(p)
                if isinstance(pval, solph.plumbing._Sequence):
                    # print(p)
                    # print(pval.highest_index)
                    pval = parse_sequence(pval)

                try:
                    if FLOW_DEFAULT[p] != pval:
                        if p == "investment":
                            # import ipdb;ipdb.set_trace()
                            flow_arguments.append(f"{p}={parse_investment(pval)}")
                        else:
                            flow_arguments.append(f"{p}={pval}")
                except ValueError:
                    flow_arguments.append(f"{p}={pval}")
            else:
                if p[0] != "_":
                    flow_arguments.append(f"{p}={pval}")
                print(p, "not in the outputs of flow")
        bus_flow_pairs.append(f"{bus.label}: solph.Flow({', '.join(flow_arguments )})")
    return bus_flow_pairs


def parse_sequence(v):
    if len(v) == 0:
        v = f"{v[0]}"
    else:
        v = f"[{','.join([str(el) for el in v])}]"
    return v


class ESCodeRenderer:
    def __init__(self, energy_system):
        self.energy_system = energy_system
        self.es_variable_name = "energy_system"
        self.busses = [n for n in self.energy_system.nodes if isinstance(n, solph.Bus)]
        self.sources = [
            n for n in self.energy_system.nodes if isinstance(n, solph.Source)
        ]
        self.sinks = [n for n in self.energy_system.nodes if isinstance(n, solph.Sink)]
        self.transformers = [
            n for n in self.energy_system.nodes if isinstance(n, solph.Transformer)
        ]
        self.storages = [
            n
            for n in self.energy_system.nodes
            if isinstance(n, solph.components.GenericStorage)
        ]
        self.chps = [
            n
            for n in self.energy_system.nodes
            if isinstance(n, solph.components.GenericCHP)
        ]
        self.extractions_turbines = [
            n
            for n in self.energy_system.nodes
            if isinstance(n, solph.components.ExtractionTurbineCHP)
        ]
        self.offset_transformers = [
            n
            for n in self.energy_system.nodes
            if isinstance(n, solph.components.OffsetTransformer)
        ]

    def print(self, fname=None):
        if fname is None:
            answer = (
                self.print_import_statements()
                + self.print_es_definition()
                + self.print_busses()
                + self.print_transformers()
                + self.print_chps()
                + self.print_extraction_turbines()
                + self.print_offet_transformers()
                + self.print_storages()
                + self.print_sources()
                + self.print_sinks()
            )
            for l in answer:
                print(l)

    def print_import_statements(self):
        answer = [""]
        answer.append("import oemof.solph as solph")
        answer.append("")
        return answer

    def print_es_definition(self):
        return ["", f"{self.es_variable_name} = solph.EnergySystem()", ""]

    def print_busses(self):
        answer = [""]
        for bus in self.busses:
            answer.append(f"{bus.label} = solph.Bus(label='{bus.label}')")
        answer.append("")
        return answer

    def print_transformers(self):
        answer = []
        for i, t in enumerate(self.transformers):
            answer.append("")
            answer = answer + self.print_single_transformer(t, variable_name=f"t{i+1}")
            answer.append("")
        return answer

    def print_single_transformer(self, t, variable_name="t"):
        all_args = {}
        all_args["label"] = f"'{t.label}'"
        input_params = parse_bus_flow_dict(t.inputs)

        all_args["inputs"] = "{" + ", ".join(input_params) + "}"

        output_params = parse_bus_flow_dict(t.outputs)

        all_args["outputs"] = "{" + ", ".join(output_params) + "}"

        conv_factors = []
        for k, v in t.conversion_factors.items():
            parse_sequence(v)
            conv_factors.append(f"{k}: {v}")
        all_args["conversion_factors"] = "{" + ", ".join(conv_factors) + "}"

        answer = []
        answer.append(f"{variable_name} = solph.Transformer(")
        for a, v in all_args.items():
            answer.append("  " + f"{a}={v},")
        answer.append(")")
        answer.append("")
        answer.append(f"{self.es_variable_name}.add({variable_name})")
        answer.append("")

        return answer

    def print_chps(self):
        answer = []
        for i, t in enumerate(self.chps):
            answer.append("")
            answer = answer + self.print_single_chp(t, variable_name=f"chp{i+1}")
            answer.append("")
        return answer

    def print_single_chp(self, t, variable_name="t"):
        all_args = {}
        all_args["label"] = f"'{t.label}'"

        for k, v in t.__dict__.items():
            if is_bus_flow_dict(v) is True:
                dict_arguments = parse_bus_flow_dict(v)
                all_args[k] = "{" + ", ".join(dict_arguments) + "}"
            elif k[0] != "_":
                all_args[k] = v

        answer = []
        answer.append(f"{variable_name} = solph.components.GenericCHP(")
        for a, v in all_args.items():
            answer.append("  " + f"{a}={v},")
        answer.append(")")
        answer.append("")
        answer.append(f"{self.es_variable_name}.add({variable_name})")
        answer.append("")

        return answer

    def print_sources(self):
        answer = []
        for i, s in enumerate(self.sources):
            answer.append("")
            answer = answer + self.print_single_source(s, variable_name=f"source{i+1}")
            answer.append("")
        return answer

    def print_single_source(self, s, variable_name="s"):
        all_args = {}
        all_args["label"] = f"'{s.label}'"

        output_params = parse_bus_flow_dict(s.outputs)
        all_args["outputs"] = "{" + ", ".join(output_params) + "}"

        answer = []
        answer.append(f"{variable_name} = solph.Source(")
        for a, v in all_args.items():
            answer.append("  " + f"{a}={v},")
        answer.append(")")
        answer.append("")
        answer.append(f"{self.es_variable_name}.add({variable_name})")
        answer.append("")

        return answer

    def print_sinks(self):
        answer = []
        for i, s in enumerate(self.sinks):
            answer.append("")
            answer = answer + self.print_single_sink(s, variable_name=f"sink{i+1}")
            answer.append("")
        return answer

    def print_single_sink(self, s, variable_name="s"):
        all_args = {}
        all_args["label"] = f"'{s.label}'"
        input_params = parse_bus_flow_dict(s.inputs)
        all_args["inputs"] = "{" + ", ".join(input_params) + "}"

        answer = []
        answer.append(f"{variable_name} = solph.Sink(")
        for a, v in all_args.items():
            answer.append("  " + f"{a}={v},")
        answer.append(")")
        answer.append("")
        answer.append(f"{self.es_variable_name}.add({variable_name})")
        answer.append("")

        return answer

    def print_storages(self):
        answer = []
        for i, s in enumerate(self.storages):
            answer.append("")
            answer = answer + self.print_single_storage(
                s, variable_name=f"storage{i+1}"
            )
            answer.append("")
        return answer

    def print_single_storage(self, s, variable_name="s"):
        return self.print_custom(
            s, variable_name, component_name="solph.component.GenericStorage"
        )

    def print_extraction_turbines(self):
        answer = []
        for i, s in enumerate(self.extractions_turbines):
            answer.append("")
            answer = answer + self.print_single_extraction_turbine(
                s, variable_name=f"extr_turb{i+1}"
            )
            answer.append("")
        return answer

    def print_single_extraction_turbine(self, s, variable_name="s"):
        return self.print_custom(
            s, variable_name, component_name="solph.component.ExtractionTurbineCHP"
        )

    def print_offet_transformers(self):
        answer = []
        for i, t in enumerate(self.offset_transformers):
            answer.append("")
            answer = answer + self.print_single_offet_transformer(
                t, variable_name=f"offset_t{i + 1}"
            )
            answer.append("")
        return answer

    def print_single_offet_transformer(self, t, variable_name="s"):
        return self.print_custom(
            t, variable_name, component_name="solph.component.OffsetTransformer"
        )

    def print_custom(self, s, variable_name="s", component_name="solph.component"):
        all_args = {}
        all_args["label"] = f"'{s.label}'"

        input_params = parse_bus_flow_dict(s.inputs)
        all_args["inputs"] = "{" + ", ".join(input_params) + "}"

        output_params = parse_bus_flow_dict(s.outputs)
        all_args["outputs"] = "{" + ", ".join(output_params) + "}"

        for k, v in s.__dict__.items():
            print(k)
            print(v)
            if is_bus_flow_dict(v) is True:
                dict_arguments = parse_bus_flow_dict(v)
                all_args[k] = "{" + ", ".join(dict_arguments) + "}"
            elif k[0] != "_":
                if isinstance(v, solph.plumbing._Sequence):
                    v = parse_sequence(v)
                    # print(k)
                    # import ipdb; ipdb.set_trace()

                all_args[k] = v

        answer = []
        answer.append(f"{variable_name} = {component_name}(")
        for a, v in all_args.items():
            answer.append("  " + f"{a}={v},")
        answer.append(")")
        answer.append("")
        answer.append(f"{self.es_variable_name}.add({variable_name})")
        answer.append("")

        return answer


# cr = ESCodeRenderer(model)
# cr.print()



"""
Module F0 - Output
==================

The model F0 output defines all functions that store evaluation results to file.
- Aggregate demand profiles to a total demand profile
- Plot all energy flows for both 14 and 365 days for each energy bus
- Store timeseries of all energy flows to excel (one sheet = one energy bus)
- Execute function: plot optimised capacities as a barchart (F1)
- Execute function: plot all annuities as a barchart (F1)
- Store scalars/KPI to excel
- Process dictionary so that it can be stored to Json
- Store dictionary to Json
"""


import json
import logging
import os

import pandas as pd

from multi_vector_simulator.B0_data_input_json import convert_from_special_types_to_json
from multi_vector_simulator.E1_process_results import get_units_of_cost_matrix_entries
import multi_vector_simulator.F1_plotting as F1_plots

try:
    import multi_vector_simulator.F2_autoreport as autoreport

    AUTOREPORT = True
except ModuleNotFoundError:
    logging.warning("The reporting feature is disabled")
    AUTOREPORT = False

from multi_vector_simulator.utils.constants import (
    SIMULATION_SETTINGS,
    PATH_OUTPUT_FOLDER,
    OUTPUT_FOLDER,
    LOGFILE,
    PATHS_TO_PLOTS,
)

from multi_vector_simulator.utils.constants import (
    JSON_WITH_RESULTS,
    JSON_FILE_EXTENSION,
)
from multi_vector_simulator.utils.constants_json_strings import (
    UNIT,
    KPI,
    OPTIMIZED_FLOWS,
    DEMANDS,
    RESOURCES,
    LES_ENERGY_VECTOR_S,
    KPI_SCALARS_DICT,
    KPI_SCALAR_MATRIX,
    KPI_COST_MATRIX,
    PROJECT_DATA,
    ECONOMIC_DATA,
    SIMULATION_RESULTS,
    LOGS,
    ERRORS,
    WARNINGS,
    FIX_COST,
    ENERGY_BUSSES,
)


def evaluate_dict(dict_values, path_pdf_report=None, path_png_figs=None):
    """This is the main function of F0. It calls all functions that prepare the simulation output, ie. Storing all simulation output into excellent files, bar charts, and graphs.

    Parameters
    ----------
    dict_values :
        dict Of all input and output parameters up to F0

    path_pdf_report : (str)
        if provided, generate a pdf report of the simulation to the given path

    path_png_figs : (str)
        if provided, generate png figures of the simulation's results to the given path

    Returns
    -------
    type
        NA

    """

    logging.info(
        "Summarizing simulation results to results_timeseries and results_scalars_assets."
    )

    parse_simulation_log(
        path_log_file=os.path.join(
            dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER], LOGFILE
        ),
        dict_values=dict_values,
    )

    # storing all flows to exel.
    store_timeseries_all_busses_to_excel(dict_values)

    # Write everything to file with multiple tabs
    store_scalars_to_excel(dict_values)

    store_as_json(
        dict_values,
        dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER],
        JSON_WITH_RESULTS,
    )

    # generate png figures
    if path_png_figs is not None:
        # plot demand timeseries
        F1_plots.plot_timeseries(
            dict_values, data_type=DEMANDS, file_path=path_png_figs
        )
        # plot demand timeseries for the first 2 weeks only
        F1_plots.plot_timeseries(
            dict_values, data_type=DEMANDS, max_days=14, file_path=path_png_figs
        )

        # plot supply timeseries
        F1_plots.plot_timeseries(
            dict_values, data_type=RESOURCES, file_path=path_png_figs
        )
        # plot supply timeseries for the first 2 weeks only
        F1_plots.plot_timeseries(
            dict_values, data_type=RESOURCES, max_days=14, file_path=path_png_figs
        )

        # plot power flows in the energy system
        F1_plots.plot_instant_power(dict_values, file_path=path_png_figs)

        # plot optimal capacities if there are optimized assets
        F1_plots.plot_optimized_capacities(dict_values, file_path=path_png_figs)

        # plot annuity, first-investment and om costs
        F1_plots.plot_piecharts_of_costs(dict_values, file_path=path_png_figs)

    # generate a pdf report
    if path_pdf_report is not None:
        app = autoreport.create_app(dict_values)
        autoreport.print_pdf(app, path_pdf_report=path_pdf_report)
        logging.info(
            "Generating PDF report of the simulation: {}".format(path_pdf_report)
        )


def store_scalars_to_excel(dict_values):
    """All output data that is a scalar is storage to an excellent file tab. This could for example be economical data or technical data.

    Parameters
    ----------
    dict_values :
        dict Of all input and output parameters up to F0

    Returns
    -------
    type
        Excel file with scalar data

    """
    results_scalar_output_file = "/scalars" + ".xlsx"

    with pd.ExcelWriter(
        dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER]
        + results_scalar_output_file
    ) as open_file:  # doctest: +SKIP
        for kpi_set in dict_values[KPI]:
            if isinstance(dict_values[KPI][kpi_set], dict):
                data = pd.DataFrame([dict_values[KPI][kpi_set]])
            else:
                data = dict_values[KPI][kpi_set]

            # Transpose results and add units to the entries
            if kpi_set == KPI_SCALARS_DICT:
                data = data.transpose()
                units_cost_kpi = get_units_of_cost_matrix_entries(
                    dict_values[ECONOMIC_DATA], dict_values[KPI][kpi_set]
                )
                data[UNIT] = units_cost_kpi

            data.to_excel(open_file, sheet_name=kpi_set)
            logging.debug(
                "Saved scalar results to: %s, tab %s.",
                results_scalar_output_file,
                kpi_set,
            )


def store_timeseries_all_busses_to_excel(dict_values):
    """This function plots the energy flows of each single bus and the energy system and saves it as PNG and additionally as a tab and an Excel sheet.

    Parameters
    ----------
    dict_values :
        dict Of all input and output parameters up to F0

    Returns
    -------
    type
        Plots and excel with all timeseries of each bus

    """

    timeseries_output_file = "/timeseries_all_busses" + ".xlsx"
    with pd.ExcelWriter(
        dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER] + timeseries_output_file
    ) as open_file:  # doctest: +SKIP
        for bus in dict_values[OPTIMIZED_FLOWS]:
            dict_values[OPTIMIZED_FLOWS][bus].to_excel(open_file, sheet_name=bus)

    logging.debug("Saved flows at busses to: %s.", timeseries_output_file)


def parse_simulation_log(path_log_file, dict_values):
    """Gather a log file with several log messages, this function gathers them all and inputs them into the dict with
    all input and output parameters up to F0

    Parameters
    ----------
    path_log_file: str/None
        path to the mvs log file
        Default: None

    dict_values :
        dict Of all input and output parameters up to F0

    Returns
    -------
    Updates the results dictionary with the log messages of the simulation

    Notes
    -----
    This function is tested with:
    - test_F0_output.TestLogCreation.test_parse_simulation_log

    """
    # Dictionaries to gather non-fatal warning and error messages that appear during the simulation

    error_dict, warning_dict = {}, {}

    if path_log_file is None:
        path_log_file = os.path.join(OUTPUT_FOLDER, LOGFILE)

    with open(path_log_file) as log_messages:
        log_messages = log_messages.readlines()

    i = j = 0

    # Loop through the list of lines of the log file to check for the relevant log messages and gather them in dicts
    for line in log_messages:
        if "ERROR" in line:
            i = i + 1
            substrings = line.split(" - ")
            message_string = substrings[-1]
            error_dict.update({i: message_string})
        elif "WARNING" in line:
            j = j + 1
            substrings = line.split(" - ")
            message_string = substrings[-1]
            warning_dict.update({j: message_string})

    log_dict = {ERRORS: error_dict, WARNINGS: warning_dict}

    dict_values[SIMULATION_RESULTS].update({LOGS: log_dict})


def store_as_json(dict_values, output_folder=None, file_name=None):
    """Converts dict_values to JSON format and saves dict_values as a JSON file or return json

    Parameters
    ----------
    dict_values : (dict)
        dict to be stored as json
    output_folder : (path)
        Folder into which json should be stored
        Default None
    file_name : (str)
        Name of the file the json should be stored as
        Default None

    Returns
    -------
    If file_name is provided, the json variable converted from the dict_values is saved under
    this file_name, otherwise the json variable is returned
    """
    json_data = json.dumps(
        dict_values,
        skipkeys=False,
        sort_keys=True,
        default=convert_from_special_types_to_json,
        indent=4,
    )
    if file_name is not None:
        file_path = os.path.abspath(os.path.join(output_folder, file_name + ".json"))

        if os.path.exists(os.path.dirname(file_path)):
            myfile = open(file_path, "w")

            myfile.write(json_data)
            myfile.close()
            logging.info(
                "Converted and stored processed simulation data to json: %s", file_path
            )
            answer = file_path
        else:
            answer = None
    else:
        answer = json_data

    return answer



r"""
Module data_parser
==================

This module defines all functions to convert formats between EPA and MVS
- Define similar parameters mapping between the EPA and MVS in MAP_EPA_MVS and MAP_MVS_EPA
- Define which fields are expected in asset list of EPA for various assets' groups in EPA_ASSET_KEYS
- Convert MVS to EPA
- Convert EPA to MVS
"""

import pprint
import logging
import json
from copy import deepcopy

from multi_vector_simulator.utils import compare_input_parameters_with_reference


from multi_vector_simulator.utils.constants import (
    MISSING_PARAMETERS_KEY,
    EXTRA_PARAMETERS_KEY,
    DATA_TYPE_JSON_KEY,
    TYPE_SERIES,
    TYPE_NONE,
    TYPE_BOOL,
    KNOWN_EXTRA_PARAMETERS,
    DEFAULT_CONSTRAINT_VALUES,
    DEFAULT_VALUE,
)

from multi_vector_simulator.utils.constants_json_strings import (
    PROJECT_DATA,
    ECONOMIC_DATA,
    SIMULATION_SETTINGS,
    CONSTRAINTS,
    ENERGY_CONSUMPTION,
    ENERGY_CONVERSION,
    ENERGY_PRODUCTION,
    ENERGY_STORAGE,
    ENERGY_BUSSES,
    ENERGY_PROVIDERS,
    UNIT,
    LABEL,
    OEMOF_ASSET_TYPE,
    ENERGY_VECTOR,
    INFLOW_DIRECTION,
    CONNECTED_CONSUMPTION_SOURCE,
    CONNECTED_FEEDIN_SINK,
    ENERGY_PRICE,
    FEEDIN_TARIFF,
    PEAK_DEMAND_PRICING,
    PEAK_DEMAND_PRICING_PERIOD,
    RENEWABLE_SHARE_DSO,
    OUTFLOW_DIRECTION,
    DEVELOPMENT_COSTS,
    DISPATCH_PRICE,
    DISPATCHABILITY,
    INSTALLED_CAP,
    LIFETIME,
    MAXIMUM_CAP,
    MAXIMUM_ADD_CAP,
    OPTIMIZE_CAP,
    OPTIMIZED_ADD_CAP,
    SPECIFIC_COSTS,
    SPECIFIC_COSTS_OM,
    SPECIFIC_REPLACEMENT_COSTS_INSTALLED,
    SPECIFIC_REPLACEMENT_COSTS_OPTIMIZED,
    TIMESERIES,
    AGE_INSTALLED,
    RENEWABLE_ASSET_BOOL,
    EFFICIENCY,
    INPUT_POWER,
    OUTPUT_POWER,
    STORAGE_CAPACITY,
    PROJECT_ID,
    PROJECT_NAME,
    SCENARIO_ID,
    SCENARIO_NAME,
    START_DATE,
    EVALUATED_PERIOD,
    OUTPUT_LP_FILE,
    MINIMAL_RENEWABLE_FACTOR,
    MINIMAL_DEGREE_OF_AUTONOMY,
    FIX_COST,
    KPI,
    TIMESTEP,
    KPI_SCALARS_DICT,
    VALUE,
    EMISSION_FACTOR,
    MAXIMUM_EMISSIONS,
    FLOW,
    KPI_UNCOUPLED_DICT,
    KPI_COST_MATRIX,
    KPI_SCALAR_MATRIX,
    SOC_INITIAL,
    SCENARIO_DESCRIPTION,
    TIMESERIES_SOC,
    TYPE_ASSET,
    DSM,
    THERM_LOSSES_REL,
    THERM_LOSSES_ABS,
    NET_ZERO_ENERGY,
    COST_REPLACEMENT,
    ASSET_DICT,
)

from multi_vector_simulator.utils.exceptions import MissingParameterError

pp = pprint.PrettyPrinter(indent=4)

MAP_EPA_MVS = {
    "economic_data": ECONOMIC_DATA,
    "energy_providers": ENERGY_PROVIDERS,
    "energy_busses": ENERGY_BUSSES,
    "energy_consumption": ENERGY_CONSUMPTION,
    "energy_conversion": ENERGY_CONVERSION,
    "energy_production": ENERGY_PRODUCTION,
    "energy_storage": ENERGY_STORAGE,
    "project_data": PROJECT_DATA,
    "simulation_settings": SIMULATION_SETTINGS,
    "energy_vector": ENERGY_VECTOR,
    "installed_capacity": INSTALLED_CAP,
    "capacity": STORAGE_CAPACITY,
    # "input_bus_name": INFLOW_DIRECTION,
    # "output_bus_name": OUTFLOW_DIRECTION,
    "input_power": INPUT_POWER,
    "output_power": OUTPUT_POWER,
    "optimize_capacity": OPTIMIZE_CAP,
    "optimized_add_cap": OPTIMIZED_ADD_CAP,
    "maximum_capacity": MAXIMUM_CAP,
    "maximum_add_cap": MAXIMUM_ADD_CAP,
    "input_timeseries": TIMESERIES,
    "constraints": CONSTRAINTS,
    "renewable_asset": RENEWABLE_ASSET_BOOL,
    KPI: KPI,
    FIX_COST: FIX_COST,
    "time_step": TIMESTEP,
    "data": VALUE,
    "replacement_costs_during_project_lifetime": COST_REPLACEMENT,
    "specific_replacement_costs_of_installed_capacity": SPECIFIC_REPLACEMENT_COSTS_INSTALLED,
    "specific_replacement_costs_of_optimized_capacity": SPECIFIC_REPLACEMENT_COSTS_OPTIMIZED,
    "asset_type": TYPE_ASSET,
    "capex_fix": DEVELOPMENT_COSTS,
    "capex_var": SPECIFIC_COSTS,
    "opex_fix": SPECIFIC_COSTS_OM,
    "opex_var": DISPATCH_PRICE,
}

MAP_MVS_EPA = {value: key for (key, value) in MAP_EPA_MVS.items()}

# Fields expected for parameters of json returned to EPA, all assets will be returned
EPA_PARAM_KEYS = {
    PROJECT_DATA: [PROJECT_ID, PROJECT_NAME, SCENARIO_ID, SCENARIO_NAME],
    SIMULATION_SETTINGS: [START_DATE, EVALUATED_PERIOD, TIMESTEP, OUTPUT_LP_FILE],
    KPI: [KPI_SCALARS_DICT, KPI_UNCOUPLED_DICT, KPI_COST_MATRIX, KPI_SCALAR_MATRIX],
}

# Fields expected for assets' parameters of json returned to EPA
EPA_ASSET_KEYS = {
    ENERGY_PROVIDERS: [
        "unique_id",
        "asset_type",
        LABEL,
        OEMOF_ASSET_TYPE,
        "energy_vector",
        INFLOW_DIRECTION,
        OUTFLOW_DIRECTION,
        CONNECTED_CONSUMPTION_SOURCE,
        CONNECTED_FEEDIN_SINK,
        DEVELOPMENT_COSTS,
        DISPATCH_PRICE,
        ENERGY_PRICE,
        FEEDIN_TARIFF,
        "installed_capacity",
        LIFETIME,
        "optimize_capacity",
        PEAK_DEMAND_PRICING,
        PEAK_DEMAND_PRICING_PERIOD,
        RENEWABLE_SHARE_DSO,
        SPECIFIC_COSTS,
        SPECIFIC_COSTS_OM,
        UNIT,
        FLOW,
    ],
    ENERGY_CONSUMPTION: [
        "unique_id",
        "asset_type",
        LABEL,
        INFLOW_DIRECTION,
        OEMOF_ASSET_TYPE,
        DEVELOPMENT_COSTS,
        DISPATCH_PRICE,
        "installed_capacity",
        LIFETIME,
        SPECIFIC_COSTS,
        SPECIFIC_COSTS_OM,
        "energy_vector",
        FLOW,
    ],
    ENERGY_CONVERSION: [
        "unique_id",
        "asset_type",
        LABEL,
        "energy_vector",
        OEMOF_ASSET_TYPE,
        INFLOW_DIRECTION,
        OUTFLOW_DIRECTION,
        AGE_INSTALLED,
        DEVELOPMENT_COSTS,
        DISPATCH_PRICE,
        EFFICIENCY,
        "installed_capacity",
        LIFETIME,
        "optimize_capacity",
        "optimized_add_cap",
        SPECIFIC_COSTS,
        SPECIFIC_COSTS_OM,
        FLOW,
    ],
    ENERGY_PRODUCTION: [
        "unique_id",
        "asset_type",
        LABEL,
        OEMOF_ASSET_TYPE,
        OUTFLOW_DIRECTION,
        DEVELOPMENT_COSTS,
        DISPATCH_PRICE,
        DISPATCHABILITY,
        "installed_capacity",
        LIFETIME,
        "maximum_capacity",
        "maximum_add_cap",
        "optimize_capacity",
        "optimized_add_cap",
        SPECIFIC_COSTS,
        SPECIFIC_COSTS_OM,
        AGE_INSTALLED,
        "renewable_asset",
        "energy_vector",
        FLOW,
    ],
    ENERGY_STORAGE: [
        "unique_id",
        "asset_type",
        LABEL,
        "energy_vector",
        INFLOW_DIRECTION,
        OUTFLOW_DIRECTION,
        OEMOF_ASSET_TYPE,
        INPUT_POWER,
        OUTPUT_POWER,
        STORAGE_CAPACITY,
        "optimize_capacity",
        "optimized_add_cap",
        TIMESERIES_SOC,
    ],
    ENERGY_BUSSES: [LABEL, "assets", "energy_vector"],
}


def convert_epa_params_to_mvs(epa_dict):
    """Convert the EPA output parameters to MVS input parameters

    Parameters
    ----------
    epa_dict: dict
        parameters from EPA user interface

    Returns
    -------
    dict_values: dict
        MVS json file, generated from EPA inputs, to be provided as MVS input

    Notes
    -----

    - For `simulation_settings`:
        - parameter `TIMESTEP` is parsed as unit-value pair
        - `OUTPUT_LP_FILE` is set to `False` by default
    - For `project_data`: parameter `SCENARIO_DESCRIPTION` is defined as placeholder string.
    - `fix_cost` is not required, default value will be set if it is not provided.
    - For missing asset group `CONSTRAINTS` following parameters are added:
        - MINIMAL_RENEWABLE_FACTOR: 0
        - MAXIMUM_EMISSIONS: None
        - MINIMAL_DEGREE_OF_AUTONOMY: 0
        - NET_ZERO_ENERGY: False
    - `ENERGY_STORAGE` assets:
        - Optimize cap written to main asset and removed from subassets
        - Units defined automatically (assumed: electricity system)
        - `SOC_INITIAL`: None
        - `THERM_LOSSES_REL`: 0
        - `THERM_LOSSES_ABS`: 0
    - If `TIMESERIES` parameter in asset dictionary: Redefine unit, value and label.
    - `ENERGY_PROVIDERS`:
        - Auto-define unit as kWh(el)
        - `INFLOW_DIRECTION=OUTFLOW_DIRECTION`
        - Default value for `EMISSION_FACTOR` added
    - `ENERGY_CONSUMPTION`:
        - `DSM` is `False`
        - `DISPATCHABILITY` is FALSE
    - `ENERGY_PRODUCTION`:
        - Default value for `EMISSION_FACTOR` added
        - `DISPATCHABILITY` is always `False`, as no dispatchable fuel assets possible right now. Must be tackeld by EPA.
     """
    epa_dict = deepcopy(epa_dict)
    dict_values = {}

    # Loop though one-dimensional energy system data (parameters directly in group)
    # Warnings for missing param_groups, will result in fatal error (except for fix_cost) as they can not be replaced with default values
    for param_group in [
        PROJECT_DATA,
        ECONOMIC_DATA,
        SIMULATION_SETTINGS,
        CONSTRAINTS,
        FIX_COST,
    ]:

        if MAP_MVS_EPA[param_group] in epa_dict:
            # Write entry of EPA to MVS json file
            dict_values[param_group] = epa_dict[MAP_MVS_EPA[param_group]]

            # convert fields names from EPA convention to MVS convention, if applicable
            keys_list = list(dict_values[param_group].keys())
            for k in keys_list:
                if k in MAP_EPA_MVS:
                    dict_values[param_group][MAP_EPA_MVS[k]] = dict_values[
                        param_group
                    ].pop(k)

            if param_group == SIMULATION_SETTINGS:
                timestep = dict_values[param_group].get(TIMESTEP)
                if timestep is not None:
                    dict_values[param_group][TIMESTEP] = {
                        UNIT: "min",
                        VALUE: timestep,
                    }
                # by default the lp file will not be outputted
                output_lp_file = dict_values[param_group].get(OUTPUT_LP_FILE)
                if output_lp_file is None:
                    dict_values[param_group][OUTPUT_LP_FILE] = {
                        UNIT: TYPE_BOOL,
                        VALUE: False,
                    }
                else:
                    dict_values[param_group][OUTPUT_LP_FILE] = {
                        UNIT: TYPE_BOOL,
                        VALUE: True if output_lp_file == "true" else False,
                    }

            if param_group == PROJECT_DATA:
                if SCENARIO_DESCRIPTION not in dict_values[param_group]:
                    dict_values[param_group][
                        SCENARIO_DESCRIPTION
                    ] = "[No scenario description available]"

        else:
            logging.warning(
                f"The parameters group '{MAP_MVS_EPA[param_group]}' is not present in the EPA parameters to be parsed into MVS json format"
            )

    # Loop through energy system asset groups and their assets
    # Logging warning message for missing asset groups, will not raise error if an asset group does not contain any assets
    for asset_group in [
        ENERGY_CONSUMPTION,
        ENERGY_CONVERSION,
        ENERGY_PRODUCTION,
        ENERGY_STORAGE,
        ENERGY_BUSSES,
        ENERGY_PROVIDERS,
    ]:
        if MAP_MVS_EPA[asset_group] in epa_dict:
            dict_asset = {}
            for asset in epa_dict[MAP_MVS_EPA[asset_group]]:

                asset_label = asset[LABEL]
                dict_asset[asset_label] = asset
                asset_keys = list(dict_asset[asset_label].keys())

                # change EPA style keys of an asset to MVS style ones
                for k in asset_keys:

                    if k in MAP_EPA_MVS:
                        dict_asset[asset_label][MAP_EPA_MVS[k]] = dict_asset[
                            asset_label
                        ].pop(k)

                    # for energy_storage there is an extra indentation level
                    if asset_group == ENERGY_STORAGE:
                        if k in (
                            MAP_MVS_EPA[STORAGE_CAPACITY],
                            MAP_MVS_EPA[INPUT_POWER],
                            MAP_MVS_EPA[OUTPUT_POWER],
                        ):
                            subasset = dict_asset[asset_label][MAP_EPA_MVS[k]]
                            subasset_keys = list(subasset.keys())

                            for sk in subasset_keys:
                                if sk in MAP_EPA_MVS:
                                    subasset[MAP_EPA_MVS[sk]] = subasset.pop(sk)

                            # add unit if not provided
                            # TODO deal with other vectors than electricity
                            if UNIT not in subasset:
                                if k == MAP_MVS_EPA[STORAGE_CAPACITY]:
                                    subasset[UNIT] = "kWh"
                                else:
                                    subasset[UNIT] = "kW"
                            # set the initial value of the state of charge to None
                            if k == MAP_MVS_EPA[STORAGE_CAPACITY]:
                                subasset[SOC_INITIAL] = {VALUE: None, UNIT: TYPE_NONE}
                                # move the optimize cap property from STORAGE_CAPACITY to the asset level
                                if OPTIMIZE_CAP in subasset:
                                    dict_asset[asset_label][
                                        OPTIMIZE_CAP
                                    ] = subasset.pop(OPTIMIZE_CAP)

                # move the unit outside the timeseries dict
                if TIMESERIES in dict_asset[asset_label]:
                    unit = dict_asset[asset_label][TIMESERIES].pop(UNIT)
                    data = dict_asset[asset_label][TIMESERIES].pop(VALUE)
                    dict_asset[asset_label][
                        UNIT
                    ] = unit  # todo this is a trick, as "UNIT" was not given
                    dict_asset[asset_label][TIMESERIES][VALUE] = data
                    dict_asset[asset_label][TIMESERIES][
                        DATA_TYPE_JSON_KEY
                    ] = TYPE_SERIES

                if asset_group == ENERGY_CONVERSION:
                    if DISPATCH_PRICE not in dict_asset[asset_label]:
                        dict_asset[asset_label].update(
                            {DISPATCH_PRICE: {VALUE: 0, UNIT: "factor"}}
                        )
                    if DEVELOPMENT_COSTS not in dict_asset[asset_label]:
                        dict_asset[asset_label].update(
                            {DEVELOPMENT_COSTS: {VALUE: 0, UNIT: "factor"}}
                        )

                # TODO remove this when change has been made on EPA side
                if asset_group == ENERGY_PRODUCTION:
                    dict_asset[asset_label].update({DISPATCHABILITY: False})

                # typically DSO
                if asset_group == ENERGY_PROVIDERS:
                    # unit is not provided, so default is kWh
                    if UNIT not in dict_asset[asset_label]:
                        dict_asset[asset_label][UNIT] = "kWh"
                    # if inflow direction is not provided, the same as outflow direction is used
                    if INFLOW_DIRECTION not in dict_asset[asset_label]:
                        dict_asset[asset_label][INFLOW_DIRECTION] = dict_asset[
                            asset_label
                        ][OUTFLOW_DIRECTION]
                    # format the energy price and feedin tariffs as timeseries
                    for asset_param in (ENERGY_PRICE, FEEDIN_TARIFF):
                        param_value = dict_asset[asset_label][asset_param][VALUE]
                        if isinstance(param_value, list):
                            dict_asset[asset_label][asset_param][VALUE] = {
                                VALUE: param_value,
                                DATA_TYPE_JSON_KEY: TYPE_SERIES
                            }

                # TODO remove this when change has been made on EPA side
                if asset_group == ENERGY_STORAGE:

                    if (
                        THERM_LOSSES_REL
                        not in dict_asset[asset_label][STORAGE_CAPACITY]
                    ):
                        dict_asset[asset_label][STORAGE_CAPACITY][THERM_LOSSES_REL] = {
                            UNIT: "factor",
                            VALUE: 0,
                        }
                    if (
                        THERM_LOSSES_ABS
                        not in dict_asset[asset_label][STORAGE_CAPACITY]
                    ):
                        dict_asset[asset_label][STORAGE_CAPACITY][THERM_LOSSES_ABS] = {
                            UNIT: "kWh",
                            VALUE: 0,
                        }

                    if OPTIMIZE_CAP not in dict_asset[asset_label]:
                        dict_asset[asset_label][OPTIMIZE_CAP] = {
                            UNIT: TYPE_BOOL,
                            VALUE: False,
                        }
                    else:
                        logging.warning(
                            "The optimized cap has been updated on EPA side so you can look for "
                            "this warning in data_parser.py and remove the warning and the 7 "
                            "lines of code above it as well"
                        )

                if asset_group == ENERGY_CONSUMPTION:
                    # DSM not used parameters, but to be sure it will be defined as False
                    if DSM not in dict_asset[asset_label]:
                        dict_asset[asset_label][DSM] = False
                    # Dispatchability of energy consumption assets always False
                    dict_asset[asset_label].update(
                        {DISPATCHABILITY: {UNIT: TYPE_BOOL, VALUE: False},}
                    )

                if asset_group == ENERGY_PRODUCTION or ENERGY_PROVIDERS:
                    # Emission factor only applicable for energy production assets and energy providers
                    if EMISSION_FACTOR not in dict_asset[asset_label]:
                        dict_asset[asset_label][EMISSION_FACTOR] = {
                            VALUE: KNOWN_EXTRA_PARAMETERS[EMISSION_FACTOR][
                                DEFAULT_VALUE
                            ]
                        }

            dict_values[asset_group] = dict_asset
        else:
            logging.info(
                f"The assets parameters '{MAP_MVS_EPA[asset_group]}' is not present in the EPA parameters to be parsed into MVS json format"
            )
            epa_dict.update({asset_group: {}})
            dict_values.update({asset_group: {}})

    # Check if all necessary input parameters are provided
    comparison = compare_input_parameters_with_reference(dict_values, set_default=True)

    # ToDo compare_input_parameters_with_reference() does not identify excess/missing parameters in the subassets of energyStorages.
    if EXTRA_PARAMETERS_KEY in comparison:
        warning_extra_parameters = "Following parameters are provided to the MVS that may be excess information: \n"
        for group in comparison[EXTRA_PARAMETERS_KEY]:
            warning_extra_parameters += f"- {group} ("
            for parameter in comparison[EXTRA_PARAMETERS_KEY][group]:
                if parameter not in [LABEL, "unique_id"]:
                    warning_extra_parameters += f"{parameter}, "
            warning_extra_parameters = warning_extra_parameters[:-2] + ") \n"
        logging.warning(warning_extra_parameters)

    if MISSING_PARAMETERS_KEY in comparison:
        error_msg = []

        missing_params = comparison[MISSING_PARAMETERS_KEY]
        if CONSTRAINTS in missing_params:

            if CONSTRAINTS not in dict_values:
                dict_values[CONSTRAINTS] = {}

            for missing_constraint in missing_params[CONSTRAINTS]:
                dict_values[CONSTRAINTS][
                    missing_constraint
                ] = DEFAULT_CONSTRAINT_VALUES[missing_constraint]

            missing_params.pop(CONSTRAINTS)

        if SIMULATION_SETTINGS in missing_params:
            if (
                OUTPUT_LP_FILE in missing_params[SIMULATION_SETTINGS]
                and len(missing_params[SIMULATION_SETTINGS]) == 1
            ):
                dict_values[SIMULATION_SETTINGS][OUTPUT_LP_FILE] = {
                    UNIT: TYPE_BOOL,
                    VALUE: False,
                }
                missing_params.pop(SIMULATION_SETTINGS)

        if FIX_COST in missing_params:
            dict_values[FIX_COST] = {}
            missing_params.pop(FIX_COST)

        error_msg.append(" ")
        error_msg.append(" ")
        error_msg.append(
            "The following parameter groups and sub parameters are missing from input parameters:"
        )

        if len(missing_params.keys()) > 0:

            for asset_group in missing_params.keys():
                # Only raise an error about missing parameter if an asset group contains assets
                if len(dict_values[asset_group].keys()) > 0:
                    error_msg.append(asset_group)
                if missing_params[asset_group] is not None:
                    for k in missing_params[asset_group]:
                        error_msg.append(f"\t`{k}` parameter")

            raise (MissingParameterError("\n".join(error_msg)))

    return dict_values


def convert_mvs_params_to_epa(mvs_dict, verbatim=False):
    """Convert the MVS output parameters to EPA format

    Parameters
    ----------
    mvs_dict: dict
        output parameters from MVS

    Returns
    -------
    epa_dict: dict
        epa parameters

    """

    epa_dict = {}

    # manage which parameters are kept and which one are removed in epa_dict
    for param_group in EPA_PARAM_KEYS:

        # translate field name from mvs to epa
        param_group_epa = MAP_MVS_EPA[param_group]

        # assign the whole MVS value to the EPA field
        epa_dict[param_group_epa] = mvs_dict[param_group]

        keys_list = list(epa_dict[param_group_epa].keys())
        for k in keys_list:
            # ditch all subfields which are not present in the EPA_PARAM_KEYS value corresponding
            # to the parameter group (except for CONSTRAINTS)
            if k not in EPA_PARAM_KEYS[param_group] or param_group in (CONSTRAINTS,):
                epa_dict[param_group_epa].pop(k)
            else:
                # convert fields names from MVS convention to EPA convention, if applicable
                if k in MAP_MVS_EPA:
                    epa_dict[param_group_epa][MAP_MVS_EPA[k]] = epa_dict[
                        param_group_epa
                    ].pop(k)

                if k == KPI_UNCOUPLED_DICT:
                    epa_dict[param_group_epa][k] = json.loads(
                        epa_dict[param_group_epa][k].to_json(orient="index")
                    )

                if k in (KPI_SCALAR_MATRIX, KPI_COST_MATRIX):

                    cols = epa_dict[param_group_epa][k].columns
                    epa_dict[param_group_epa][k].columns = [
                        MAP_MVS_EPA.get(k, k) for k in cols
                    ]
                    epa_dict[param_group_epa][k] = json.loads(
                        epa_dict[param_group_epa][k]
                        .set_index("label")
                        .to_json(orient="index")
                    )

                # if the parameter is of type
                if k == OUTPUT_LP_FILE:
                    if epa_dict[param_group_epa][k][UNIT] == TYPE_BOOL:
                        epa_dict[param_group_epa].pop(k)

    # manage which assets parameters are kept and which one are removed in epa_dict
    for asset_group in EPA_ASSET_KEYS:
        list_asset = []
        for asset_label in mvs_dict[asset_group]:
            # mvs[asset_group] is a dict we want to change into a list

            # each asset is also a dict
            asset = mvs_dict[asset_group][asset_label]

            # if the asset possesses a unit field
            if UNIT in asset:
                unit = asset.pop(UNIT)
            else:
                unit = None

            unit_soc = None

            # keep the information about the dict key, but move it into the dict value
            asset[LABEL] = asset_label

            asset_keys = list(asset.keys())
            for k in asset_keys:
                if k in MAP_MVS_EPA:
                    # convert some keys MVS to EPA style according to the mapping
                    asset[MAP_MVS_EPA[k]] = asset.pop(k)
                # TODO change energy busses from dict to list in MVS
                if asset_group == ENERGY_BUSSES and k == ASSET_DICT:
                    asset["assets"] = list(asset.pop(k).keys())
                if asset_group == ENERGY_STORAGE:
                    if k in (INPUT_POWER, OUTPUT_POWER, STORAGE_CAPACITY):
                        asset[k] = mvs_dict[asset_group][asset_label][MAP_MVS_EPA[k]]
                        subasset_keys = list(asset[k].keys())

                        # if the asset possesses a unit field
                        if UNIT in asset[k]:
                            subunit = asset[k].pop(UNIT)
                            if k == STORAGE_CAPACITY:
                                unit_soc = subunit
                        else:
                            subunit = None

                        for sk in subasset_keys:
                            if sk in MAP_MVS_EPA:
                                # convert some keys MVS to EPA style according to the mapping
                                asset[k][MAP_MVS_EPA[sk]] = asset[k].pop(sk)
                        # convert pandas.Series to a timeseries dict with key DATA value list,
                        # move the unit inside the timeseries dict under key UNIT
                        if FLOW in asset[k]:
                            timeseries = asset[k][FLOW].to_list()
                            asset[k][FLOW] = {UNIT: subunit, VALUE: timeseries}

            if MAP_MVS_EPA[TIMESERIES] in asset:
                asset.pop(MAP_MVS_EPA[TIMESERIES])

            # convert pandas.Series to a timeseries dict with key DATA value list,
            # move the unit inside the timeseries dict under key UNIT
            if FLOW in asset:
                if isinstance(asset.get(OUTFLOW_DIRECTION, None), list):
                    timeseries = {}
                    for bus in asset[OUTFLOW_DIRECTION]:
                        timeseries[bus] = asset[FLOW][bus].to_list()
                    asset[FLOW] = {UNIT: unit, VALUE: timeseries}
                else:
                    timeseries = asset[FLOW].to_list()
                    asset[FLOW] = {UNIT: unit, VALUE: timeseries}

            if TIMESERIES_SOC in asset:
                timeseries = asset[TIMESERIES_SOC].to_list()
                asset[TIMESERIES_SOC] = {UNIT: unit_soc, VALUE: timeseries}

            # Excess sinks should not be provided to the EPA
            if "_excess" not in asset_label:
                list_asset.append(asset)

        epa_dict[MAP_MVS_EPA[asset_group]] = list_asset

    # verify that there are extra keys, besides the one expected by EPA data structure
    extra_keys = {}
    # verify that there are keys expected by the EPA which are not filled
    missing_keys = {}
    for asset_group in EPA_ASSET_KEYS:
        extra_keys[asset_group] = []
        missing_keys[asset_group] = []
        for asset in epa_dict[MAP_MVS_EPA[asset_group]]:
            asset_keys = list(asset.keys())
            # loop over the actual fields of the asset
            for k in asset_keys:
                # remove any field which is not listed under the asset_group in EPA_ASSET_KEYS
                if k not in EPA_ASSET_KEYS[asset_group]:
                    asset.pop(k)
                    # keep trace of this extra key
                    if k not in extra_keys[asset_group]:
                        extra_keys[asset_group].append((asset[LABEL], k))
            # loop over the expected fields of the asset_group in EPA_ASSET_KEYS
            for k in EPA_ASSET_KEYS[asset_group]:
                # if a field is missing in the actual asset, keep trace of it
                if k not in asset:
                    missing_keys[asset_group].append((asset[LABEL], k))

    if verbatim is True:
        print("#" * 10 + " Missing values " + "#" * 10)
        pp.pprint(missing_keys)

        print("#" * 10 + " Extra values " + "#" * 12)
        pp.pprint(extra_keys)

    return epa_dict



import os
import copy
import json
import shutil
import logging
import warnings
import pandas as pd
from .constants import (
    PACKAGE_DATA_PATH,
    JSON_FNAME,
    CSV_ELEMENTS,
    OUTPUT_FOLDER,
    JSON_EXT,
    CSV_EXT,
    REQUIRED_MVS_PARAMETERS,
    KNOWN_EXTRA_PARAMETERS,
    JSON_FNAME,
    MISSING_PARAMETERS_KEY,
    EXTRA_PARAMETERS_KEY,
    TEMPLATE_INPUT_FOLDER,
    DEFAULT_VALUE,
    WARNING_TEXT,
)

from .constants_json_strings import (
    UNIT,
    VALUE,
    PROJECT_DATA,
    ECONOMIC_DATA,
    SIMULATION_SETTINGS,
    CONSTRAINTS,
    ENERGY_CONSUMPTION,
    ENERGY_CONVERSION,
    ENERGY_PRODUCTION,
    ENERGY_STORAGE,
    ENERGY_BUSSES,
    ENERGY_PROVIDERS,
    FIX_COST,
)

from .exceptions import MissingParameterError


class ParameterDocumentation:
    """Helper to access a parameter's information given its variable name"""

    def __init__(
        self, param_info_file, label_header="label",
    ):
        self.param_doc = pd.read_csv(param_info_file).set_index(label_header)
        self.label_hdr = label_header
        self.fname = param_info_file
        self.param_format = {"numeric": float, "str": str, "boolean": bool}

    @property
    def where_to_find_param_documentation(self):
        return (
            "*" * 5
            + f" Note: The documentation about each of the MVS parameters can be found in the csv file {self.fname}. "
            + "*" * 5
        )

    def __get_doc_parameter_info(self, param_label, column_name):
        """Search the value of a parameter information in the parameter doc

        Parameters
        ----------
        param_label: str
            name of the variable as referenced in the column "label" of the
            documentation csv file
        column_name:
            name of the documentation csv file's column corresponding to the
            desired information about the parameter

        Returns
        -------
        str: value of the given parameter information
        """
        if isinstance(param_label, list):
            answer = []
            for p_name in param_label:
                answer.append(self.__get_doc_parameter_info(p_name, column_name))
        else:
            try:
                answer = self.param_doc.loc[param_label][column_name]
            except KeyError as e:
                raise KeyError(
                    f"Either {param_label} is not part of the {self.label_hdr} column of the file {self.fname}, or the column {column_name} does not exist in this file"
                ).with_traceback(e.__traceback__)
        return answer

    def get_doc_verbose(self, param_label):
        answer = self.__get_doc_parameter_info(param_label, "verbose")
        answer_is_list = True
        if not isinstance(param_label, list):
            answer_is_list = False
            answer = [answer]
            param_label = [param_label]

        for i in range(len(answer)):
            if answer[i] == "None":
                answer[i] = param_label[i].replace("_", " ").title()
        if answer_is_list is False:
            answer = answer[0]
        return answer

    def get_doc_definition(self, param_label):
        return self.__get_doc_parameter_info(param_label, ":Definition:")

    def get_doc_default(self, param_label):
        answer = self.__get_doc_parameter_info(param_label, ":Default:")
        param_type = self.get_doc_type(param_label)
        if answer == "None":
            answer = None
        else:
            answer = self.param_format[param_type](answer)
        return answer

    def get_doc_unit(self, param_label):
        return self.__get_doc_parameter_info(param_label, ":Unit:")

    def get_doc_type(self, param_label):
        return self.__get_doc_parameter_info(param_label, ":Type:")


try:
    mvs_parameter_file = "MVS_parameters_list.csv"
    DOC_PATH = os.path.join(
        os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        ),
        "docs",
    )
    if os.path.exists(DOC_PATH):
        PARAMETERS_DOC = ParameterDocumentation(
            param_info_file=os.path.join(DOC_PATH, mvs_parameter_file)
        )
    elif os.path.exists(PACKAGE_DATA_PATH):
        PARAMETERS_DOC = ParameterDocumentation(
            param_info_file=os.path.join(PACKAGE_DATA_PATH, mvs_parameter_file)
        )
    else:
        PARAMETERS_DOC = None

except FileNotFoundError:
    PARAMETERS_DOC = None


def find_json_input_folders(
    path, specific_file_name=JSON_FNAME, ignore_folders=(OUTPUT_FOLDER,)
):
    """Recursively look in the folder structure until is sees a specific file

    Parameters
    ----------
    path: str
        the starting point of the search in the folder structure
    specific_file_name: str
        the name of the special file which should be present within a folder to add this folder
        name to the list of matching folders
    ignore_folders: tuple of str
        a tuple of folder names which should not be investigated by the function, nor added to
        the list of matching folders

    Returns
    -------
    A list of paths to folders containing a specific file
    """
    folder_list = [
        fn.name
        for fn in os.scandir(path)
        if fn.is_dir() and fn.name not in ignore_folders
    ]

    if os.path.exists(os.path.join(path, JSON_FNAME)):
        answer = [path]
    else:
        answer = []
        for folder in folder_list:
            answer = answer + find_json_input_folders(
                os.path.join(path, folder),
                specific_file_name=specific_file_name,
                ignore_folders=ignore_folders,
            )
    return answer


def find_csv_input_folders(
    path, specific_folder_name=CSV_ELEMENTS, ignore_folders=(OUTPUT_FOLDER,)
):
    """Recursively look in the folder structure until is sees a specific folder

    Parameters
    ----------
    path: str
        the starting point of the search in the folder structure
    specific_folder_name: str
        the name of the special folder which should be present within a folder to add this folder
        name to the list of matching folders
    ignore_folders: tuple of str
        a tuple of folder names which should not be investigated by the function, nor added to
        the list of matching folders

    Returns
    -------
    A list of paths to folders containing a specific folder
    """

    folder_list = [
        fn.name
        for fn in os.scandir(path)
        if fn.is_dir() and fn.name not in ignore_folders
    ]
    if CSV_ELEMENTS in folder_list:
        return [path]
    else:
        answer = []
        for folder in folder_list:
            answer = answer + find_csv_input_folders(
                os.path.join(path, folder), specific_folder_name=specific_folder_name
            )
        return answer


def compare_input_parameters_with_reference(
    folder_path, ext=JSON_EXT, flag_missing=False, set_default=False
):
    """Compare provided MVS input parameters with the required parameters

    Extra parameters listed in KNOWN_EXTRA_PARAMETERS are not flagged as missing if
    not provided, instead they take their default value defined in KNOWN_EXTRA_PARAMETERS

    Parameters
    ----------
    folder_path: str
        path to the mvs input folder
    ext: str
        one of {JSON_EXT} or {CSV_EXT}
    flag_missing: bool
        if True, raise MissingParameterError for each missing required parameter
    set_default: bool
        if True, set the default value of a missing required parameter which is listed in
        KNOWN_EXTRA_PARAMETERS

    Returns
    -------
    A dict with the missing parameters under the key {MISSING_PARAMETERS_KEY} and extra parameters
    under the key {EXTRA_PARAMETERS_KEY}

    Notes
    -----
    #todo This function does not check whether the storage-subassets have all necessary parameters, or if excess parameters are provided.
    """
    if ext == JSON_EXT:

        if isinstance(folder_path, dict):
            # the folder_path argument is already a dict
            main_parameters = folder_path
        else:
            # load the mvs input json file into a dict
            json_file_path = os.path.join(folder_path, JSON_FNAME)
            with open(json_file_path) as fp:
                main_parameters = json.load(fp)

    elif ext == CSV_EXT:
        # list the mvs input csv files
        folder_csv_path = os.path.join(folder_path, CSV_ELEMENTS)
        main_parameters = [
            fn[:-4] for fn in os.listdir(folder_csv_path) if fn.endswith(".csv")
        ]

    extra_parameters = {}
    missing_parameters = {}

    required_parameters = REQUIRED_MVS_PARAMETERS[ext]

    for mp in main_parameters:

        if mp not in required_parameters.keys():
            # the main parameter is provided but is not required --> extra
            extra_parameters[mp] = []
        else:
            # the main parameter is provided and required
            # --> comparison of the sub parameters with the reference
            if ext == JSON_EXT:
                # get the sub parameters from the json structure
                if mp in [
                    PROJECT_DATA,
                    ECONOMIC_DATA,
                    SIMULATION_SETTINGS,
                    CONSTRAINTS,
                    FIX_COST,
                ]:
                    # project parameters only
                    sub_params = {"non-asset": main_parameters[mp].keys()}
                elif mp in [
                    ENERGY_CONSUMPTION,
                    ENERGY_CONVERSION,
                    ENERGY_PRODUCTION,
                    ENERGY_STORAGE,
                    ENERGY_BUSSES,
                    ENERGY_PROVIDERS,
                ]:
                    # dict containing assets
                    # TODO this should be modified if the assets are inside a list instead
                    # of inside a dict
                    sub_params = main_parameters[mp]

            elif ext == CSV_EXT:
                # read the csv file, each line corresponds to a sub_parameter
                df = pd.read_csv(os.path.join(folder_csv_path, mp + ".csv"))
                sub_params = {"non-asset": df.iloc[:, 0].unique().tolist()}

            for k in sub_params:
                sub_parameters = sub_params[k]
                if required_parameters[mp] is not None:
                    # intersect the set of provided sub_parameters with the set of required sub parameters
                    not_matching_params = list(
                        set(sub_parameters) ^ set(required_parameters[mp])
                    )
                else:
                    # the parameter is expected to contain user defined names --> those are not checked
                    not_matching_params = []

                for sp in not_matching_params:
                    if sp in required_parameters[mp]:

                        if sp in KNOWN_EXTRA_PARAMETERS and set_default is True:
                            # the sub parameter is not provided but is listed in known extra parameters
                            # --> default value is set for this parameter
                            if k == "non-asset":
                                main_parameters[mp][sp] = {
                                    UNIT: KNOWN_EXTRA_PARAMETERS[sp][UNIT],
                                    VALUE: KNOWN_EXTRA_PARAMETERS[sp][DEFAULT_VALUE],
                                }
                                logging.warning(
                                    f"You are not using the parameter '{sp}' for asset group '{mp}', which "
                                    + KNOWN_EXTRA_PARAMETERS[sp][WARNING_TEXT]
                                    + f"This parameter is set to it's default value ({KNOWN_EXTRA_PARAMETERS[sp][DEFAULT_VALUE]}),"
                                    + " which can influence the results."
                                    + "In the next release, this parameter will required."
                                )
                            else:
                                main_parameters[mp][k][sp] = {
                                    UNIT: KNOWN_EXTRA_PARAMETERS[sp][UNIT],
                                    VALUE: KNOWN_EXTRA_PARAMETERS[sp][DEFAULT_VALUE],
                                }
                                logging.warning(
                                    f"You are not using the parameter '{sp}' for asset '{k}' of asset group '{mp}', which "
                                    + KNOWN_EXTRA_PARAMETERS[sp][WARNING_TEXT]
                                    + f"This parameter is set to it's default value ({KNOWN_EXTRA_PARAMETERS[sp][DEFAULT_VALUE]}),"
                                    + " which can influence the results."
                                    + "In the next release, this parameter will required."
                                )
                        elif set_default is True:

                            main_parameters[mp][k][sp] = {
                                VALUE: PARAMETERS_DOC.get_doc_default(sp),
                                UNIT: PARAMETERS_DOC.get_doc_unit(sp),
                            }
                            logging.warning(
                                f"You are not providing a value for the parameter '{sp}' of asset '{k}' in asset group '{mp}'"
                                + f"This parameter is then set to it's default value ({PARAMETERS_DOC.get_doc_default(sp)}).\n"
                                + PARAMETERS_DOC.where_to_find_param_documentation
                            )
                        else:
                            # the sub parameter is not provided but is required --> missing
                            param_list = missing_parameters.get(mp, [])
                            param_list.append(sp)
                            missing_parameters[mp] = param_list
                    else:
                        # the sub parameter is provided but is not required --> extra
                        param_list = extra_parameters.get(mp, [])
                        if sp not in param_list:
                            param_list.append(sp)
                            extra_parameters[mp] = param_list

    for mp in required_parameters.keys():
        if mp not in main_parameters:
            # the main parameter is not provided but is required --> missing
            missing_parameters[mp] = required_parameters[mp]

    answer = {}
    if len(missing_parameters) > 0:
        answer[MISSING_PARAMETERS_KEY] = missing_parameters
    if len(extra_parameters) > 0:
        answer[EXTRA_PARAMETERS_KEY] = extra_parameters

    if flag_missing is True:
        warn_missing_parameters(answer)

    return answer


def warn_missing_parameters(comparison_with_reference):
    """Raise error for missing parameters

    Parameters
    ----------
    comparison_with_reference: dict
        dict with possibly two keys: MISSING_PARAMETERS_KEY, EXTRA_PARAMETERS_KEY
    Returns
    -------
    Nothing

    """
    if MISSING_PARAMETERS_KEY in comparison_with_reference:
        error_msg = []

        d = comparison_with_reference[MISSING_PARAMETERS_KEY]

        error_msg.append(" ")
        error_msg.append(" ")
        error_msg.append(
            "The following parameter groups and sub parameters are missing from input parameters:"
        )

        for asset_group in d.keys():
            error_msg.append(asset_group)
            if d[asset_group] is not None:
                for k in d[asset_group]:
                    error_msg.append(f"\t`{k}` parameter")

        raise (MissingParameterError("\n".join(error_msg)))


def set_nested_value(dct, value, keys):
    r"""Set a value within a nested dict structure given the path within the dict

    Parameters
    ----------
    dct: dict
        the (potentially nested) dict from which we want to get a value
    value: variable type
        value to assign within the dict
    keys: tuple
        Tuple containing the succession of keys which lead to the value within the nested dict

    Returns
    -------
    The value under the path within the (potentially nested) dict

    Example
    -------
    >>> dct = dict(a=dict(a1=1, a2=2),b=dict(b1=dict(b11=11,b12=dict(b121=121))))
    >>> print(set_nested_value(dct, 400,("b", "b1", "b12","b121")))
    {'a': {'a1': 1, 'a2': 2}, 'b': {'b1': {'b11': 11, 'b12': {'b121': 400}}}}
    """
    if isinstance(keys, tuple) is True:
        try:
            answer = copy.deepcopy(dct)
            if keys[0] not in answer:
                raise KeyError(
                    ": pathError: that path does not exist in the nested dict"
                )
            if len(keys) > 1:
                answer[keys[0]] = set_nested_value(dct[keys[0]], value, keys[1:])
            elif len(keys) == 1:
                # if the value is a dict with structure {VALUE: ..., UNIT: ...}
                if isinstance(answer[keys[0]], dict):
                    if VALUE in answer[keys[0]]:
                        answer[keys[0]][VALUE] = value
                    else:
                        answer[keys[0]] = value
                else:
                    answer[keys[0]] = value
            else:
                raise ValueError(
                    "The tuple argument 'keys' from set_nested_value() should not be empty"
                )
        except KeyError as e:
            if "pathError" in str(e):
                raise KeyError(keys[0] + ", " + e.args[0])
    elif isinstance(keys, str) is True:
        return set_nested_value(dct, value, split_nested_path(keys))
    else:
        raise TypeError("The argument 'keys' from set_nested_value() should be a tuple")
    return answer


def get_nested_value(dct, keys):
    r"""Get a value from a succession of keys within a nested dict structure

    Parameters
    ----------
    dct: dict
        the (potentially nested) dict from which we want to get a value
    keys: tuple
        Tuple containing the succession of keys which lead to the value within the nested dict

    Returns
    -------
    The value under the path within the (potentially nested) dict

    Example
    -------
    >>> dct = dict(a=dict(a1=1, a2=2),b=dict(b1=dict(b11=11,b12=dict(b121=121))))
    >>> print(get_nested_value(dct, ("b", "b1", "b12","b121")))
    121
    """
    if isinstance(keys, tuple) is True:
        try:
            if len(keys) > 1:
                answer = get_nested_value(dct[keys[0]], keys[1:])
            elif len(keys) == 1:
                answer = dct[keys[0]]
            else:
                raise ValueError(
                    "The tuple argument 'keys' from get_nested_value() should not be empty"
                )
        except KeyError as e:
            if "pathError" in str(e):
                raise KeyError(str(keys[0]) + ", " + str(e))
            else:
                raise KeyError(
                    str(keys[0])
                    + ": pathError: that path does not exist in the nested dict"
                )

    else:
        raise TypeError("The argument 'keys' from get_nested_value() should be a tuple")
    return answer


def split_nested_path(path):
    r"""Separate a single-string path in a nested dict in a list of keys


    Parameters
    ----------
    path: str or tuple
        path within a nested dict which is expressed as a str of a succession of keys separated by
        a `.` or a `,`. The order of keys is to be read from left to right.

    Returns
    -------
    Tuple containing the succession of keys which lead to the value within the nested dict

    """
    SEPARATORS = (".", ",")
    keys_list = None
    if isinstance(path, str):
        separator_count = 0
        keys_separator = None
        for separator in SEPARATORS:
            if separator in path:
                if path.count(separator) > 0:
                    if separator_count > 0:
                        raise ValueError(
                            f"The separator of the nested dict's path is not unique"
                        )
                    separator_count = path.count(separator)
                    keys_separator = separator
        if keys_separator is not None:
            keys_list = tuple(path.split(keys_separator))
    elif isinstance(path, tuple):
        keys_list = path
    else:
        raise TypeError("The argument path is not tuple or str type")

    return keys_list


def nested_dict_crawler(dct, path=None, path_dct=None):
    r"""A recursive algorithm that crawls through a (nested) dictionary and returns
    a dictionary of endkeys mapped to a list of the paths leading to each endkey.
    An endkey is defined the last key in the nested dict before a value of type different than dict or the last key
    before a dict containing only {"unit": ..., "value": ...}
            Parameters
            ----------
            dct: dict
                the (potentially nested) dict from which we want to get the endkeys
            path: list
                storing the current path that the algorithm is on
            path_dct: dict
                result dictionary where each key is assigned to its (multiple) paths within the (nested) dictionary
            Returns
            -------
            Dictionary of key and paths to the respective key within the nested dictionary structure
            Example
            -------
            >>> dct = dict(a=dict(a1=1, a2=2),b=dict(b1=dict(b11=11,b12=dict(b121=121))))
            >>> nested_dict_crawler(dct)
            {
                "a1": [("a", "a1")],
                "a2": [("a", "a2")],
                "b11": [("b", "b1", "b11")],
                "b121": [("b", "b1", "b12", "b121")],
            }
    """
    if path is None:
        path = []
    if path_dct is None:
        path_dct = dict()

    for key, value in dct.items():
        path.append(key)
        if isinstance(value, dict):
            if "value" in value.keys() and "unit" in value.keys():
                if path[-1] in path_dct:
                    path_dct[path[-1]].append(tuple(path))
                else:
                    path_dct[path[-1]] = [tuple(path)]
            else:
                nested_dict_crawler(value, path, path_dct)
        else:
            if path[-1] in path_dct:
                path_dct[path[-1]].append(tuple(path))
            else:
                path_dct[path[-1]] = [tuple(path)]
        path.pop()
    return path_dct


def copy_report_assets(path_destination_folder):
    """Copy the css file and eland logo to the path_destination_folder

    Parameters
    ----------
    path_destination_folder: str
        path where the default report css files and logo should be copied

    Returns
    -------
    Path of the destination folder

    """
    assets_folder = os.path.join(path_destination_folder, "report", "assets")
    if os.path.exists(assets_folder) is False:
        # copy from the default asset folder from mvs package
        try:
            assets_folder = shutil.copytree(
                os.path.join(PACKAGE_DATA_PATH, "assets"), assets_folder
            )
        except FileNotFoundError:
            assets_folder = shutil.copytree(
                os.path.join(".", "report", "assets"), assets_folder
            )
    else:
        logging.warning(
            "The assets folder {} exists already, it will not be replaced by default folder "
            "from multi_vector_simulator's package".format(assets_folder)
        )
    return assets_folder


def copy_inputs_template(path_destination_folder):
    """Copy the inputs template folder

    Parameters
    ----------
    path_destination_folder: str
        path where the inputs template folder should be copied to

    Returns
    -------
    Path of the destination folder

    """
    inputs_template_folder = os.path.join(
        path_destination_folder, TEMPLATE_INPUT_FOLDER
    )
    if os.path.exists(inputs_template_folder) is False:
        # copy from the default asset folder from mvs package
        try:
            inputs_template_folder = shutil.copytree(
                os.path.join(PACKAGE_DATA_PATH, TEMPLATE_INPUT_FOLDER),
                inputs_template_folder,
            )
            logging.info(
                f"The following folder was successfully created into your local "
                f"directory {inputs_template_folder}"
            )
        except FileNotFoundError:
            logging.warning(
                "If you installed the package in develop mode, then you cannot use this command"
            )
    else:
        logging.warning(
            "The inputs template folder {} exists already, it will not be replaced by default "
            "folder from multi_vector_simulator's package".format(TEMPLATE_INPUT_FOLDER)
        )
    return inputs_template_folder



"""
Module B0 - Data input json
===========================
"""
import logging
import copy
import json
import os

import numpy as np
import pandas as pd

from multi_vector_simulator.utils import (
    data_parser,
    compare_input_parameters_with_reference,
)

from multi_vector_simulator.utils.constants_json_strings import (
    START_DATE,
    PERIODS,
    END_DATE,
    EVALUATED_PERIOD,
    TIME_INDEX,
    TIMESTEP,
    UNIT_MINUTE,
    VALUE,
    DATA,
    TIMESERIES,
)

from multi_vector_simulator.utils.constants import (
    CSV_FNAME,
    INPUTS_COPY,
    PATHS_TO_PLOTS,
    DICT_PLOTS,
    DATA_TYPE_JSON_KEY,
    TYPE_DATETIMEINDEX,
    TYPE_SERIES,
    TYPE_NDARRAY,
    TYPE_DATAFRAME,
    TYPE_TIMESTAMP,
    SIMULATION_SETTINGS,
    PATH_INPUT_FOLDER,
    PATH_OUTPUT_FOLDER,
    PATH_OUTPUT_FOLDER_INPUTS,
    MISSING_PARAMETERS_KEY,
)

"""
This module is used to open a json file and parse it as a dict all input parameters for the energy 
system.

If the user does not give the input parameters "path_input_folder", "path_output_folder" or "path_output_folder_inputs", they are replaced by the default values.

It will be an interface to the EPA.
"""


def convert_from_json_to_special_types(a_dict, prev_key=None, time_index=None):
    """Convert the field values of the mvs result json file which are not simple types.

    The function is recursive to explore all nested levels

    Parameters
    ----------
    a_dict: variable
        In the recursion, this is either a dict (moving down one nesting level) or a field value
    prev_key: str
        The previous key of the dict in the recursive loop

    Returns
    -------
    The original dictionary, with the serialized instances of pandas.Series,
    pandas.DatetimeIndex, pandas.DataFrame, numpy.array converted back to their original form

    """

    answer = a_dict
    if isinstance(a_dict, dict):
        # the a_dict argument is a dictionary not containing the special type key,
        # therefore we dive deeper in the nesting level
        if DATA_TYPE_JSON_KEY not in a_dict:
            answer = {}
            for k in a_dict:
                answer[k] = convert_from_json_to_special_types(
                    a_dict[k], prev_key=k, time_index=time_index
                )
        # TODO this cas might be obsolete with the newer version of the parser from PR #675
        elif prev_key == data_parser.MAP_MVS_EPA[TIMESERIES]:
            # the a_dict is from the EPA
            answer = pd.Series(a_dict[DATA])
            # Set time_index to Series
            if time_index is not None:
                if len(answer.index) > len(time_index):
                    logging.warning(
                        f"The time index inferred from {SIMULATION_SETTINGS} is longer as "
                        f"the timeserie under the field {prev_key}"
                    )
                elif len(answer.index) < len(time_index):
                    logging.warning(
                        f"The time index inferred from {SIMULATION_SETTINGS} is shorter as "
                        f"the timeserie under the field {prev_key}"
                    )
                else:
                    answer.index = time_index

        else:
            # the a_dict is a dictionary containing the special type key,
            # therefore we apply the conversion if this type is listed below

            # find the special type value
            data_type = a_dict.pop(DATA_TYPE_JSON_KEY)

            if TYPE_DATAFRAME in data_type:
                # pandas.DataFrame
                a_dict = json.dumps(a_dict)
                answer = pd.read_json(a_dict, orient="split")
            elif TYPE_DATETIMEINDEX in data_type:
                # pandas.DatetimeIndex
                if time_index is not None:
                    answer = time_index
                else:
                    answer = pd.DatetimeIndex(a_dict.get(VALUE, []))

                answer.freq = answer.inferred_freq
            elif TYPE_SERIES in data_type:
                # pandas.Series
                # extract the name of the series in case it was a tuple
                name = a_dict.get("name", None)

                # reconvert the dict to a json for conversion to pandas Series
                answer = pd.Series(a_dict[VALUE])

                # Set time_index to Series
                if time_index is not None:
                    if len(answer.index) > len(time_index):
                        logging.warning(
                            f"The time index inferred from {SIMULATION_SETTINGS} is shorter as "
                            f"the timeserie under the field {prev_key} ({len(time_index)}<{len(answer.index)})"
                        )
                    elif len(answer.index) < len(time_index):
                        logging.warning(
                            f"The time index inferred from {SIMULATION_SETTINGS} is longer as "
                            f"the timeserie under the field {prev_key} ({len(time_index)}>{len(answer.index)})"
                        )
                    else:
                        answer.index = time_index

                # if the name was a tuple it was converted to a list via json serialization
                if isinstance(name, list):
                    name[0] = tuple(name[0])
                    name = tuple(name)

                if name is not None:
                    answer.name = name

            elif TYPE_TIMESTAMP in data_type:
                answer = pd.Timestamp(a_dict[VALUE])
            elif TYPE_NDARRAY in data_type:
                # numpy.array
                answer = np.array(a_dict[VALUE])

    return answer


def convert_from_special_types_to_json(o):
    """This converts all data stored in dict_values that is not compatible with the
    json format to a format that is compatible.

    Parameters
    ----------
    o :
        Any type. Object to be converted to json-storable value.

    Returns
    -------
    type
        json-storable value.

    """
    if isinstance(o, np.int64) or isinstance(o, np.int32):
        answer = int(o)
    elif isinstance(o, bool) or isinstance(o, str):
        answer = o
    elif isinstance(o, pd.DatetimeIndex):
        answer = {DATA_TYPE_JSON_KEY: TYPE_DATETIMEINDEX, VALUE: o.values.tolist()}
    elif isinstance(o, pd.Timestamp):
        answer = {DATA_TYPE_JSON_KEY: TYPE_TIMESTAMP, VALUE: str(o)}
    elif isinstance(o, pd.Series):
        answer = {DATA_TYPE_JSON_KEY: TYPE_SERIES, VALUE: o.to_list()}
    elif isinstance(o, np.ndarray):
        answer = {DATA_TYPE_JSON_KEY: TYPE_NDARRAY, VALUE: o.tolist()}
    elif isinstance(o, pd.DataFrame):
        answer = {DATA_TYPE_JSON_KEY: TYPE_DATAFRAME}
        answer.update(json.loads(o.to_json(orient="split")))
    else:
        raise TypeError(
            "An error occurred when converting the simulation data (dict_values) to json, as the type is not recognized: \n"
            "Type: " + str(type(o)) + " \n "
            "Value(s): " + str(o) + "\n"
            "Please edit function CO_data_processing.dataprocessing.store_as_json."
        )

    return answer


def retrieve_date_time_info(simulation_settings):
    """
    Updates simulation settings by all time-related parameters.
    - START_DATE
    - END_DATE
    - TIME_INDEX
    - PERIODS

    Parameters
    ----------
    simulation_settings: dict
        Simulation parameters of the input data

    Returns
    -------
    Update simulation_settings by start date, end date, timeindex, and number of simulation periods


    Notes
    -----
    Function tested with test_retrieve_datetimeindex_for_simulation()
    """
    simulation_settings.update(
        {START_DATE: pd.to_datetime(simulation_settings[START_DATE])}
    )
    simulation_settings.update(
        {
            END_DATE: simulation_settings[START_DATE]
            + pd.DateOffset(days=simulation_settings[EVALUATED_PERIOD][VALUE], hours=-1)
        }
    )
    # create time index used for initializing oemof simulation
    simulation_settings.update(
        {
            TIME_INDEX: pd.date_range(
                start=simulation_settings[START_DATE],
                end=simulation_settings[END_DATE],
                freq=str(simulation_settings[TIMESTEP][VALUE]) + UNIT_MINUTE,
            )
        }
    )

    simulation_settings.update({PERIODS: len(simulation_settings[TIME_INDEX])})


def load_json(
    path_input_file,
    path_input_folder=None,
    path_output_folder=None,
    move_copy=False,
    flag_missing_values=True,
    set_default_values=False,
):
    """Opens and reads json input file and parses it to dict of input parameters.

    Parameters
    ----------

    path_input_file: str
        The path to the json file created from csv files
    path_input_folder : str, optional
        The path to the directory where the input CSVs/JSON files are located.
        Default: 'inputs/'.
    path_output_folder : str, optional
        The path to the directory where the results of the simulation such as
        the plots, time series, results JSON files are saved by MVS E-Lands.
        Default: 'MVS_outputs/'
    move_copy: bool, optional
        if this is set to True, the path_input_file will be moved to the path_output_folder
        Default: False
    flag_missing_values: bool
        if True, raise MissingParameterError for each missing required parameter
    set_default_values: bool
        if True, set the default value of a missing required parameter which is listed in
        KNOWN_EXTRA_PARAMETERS


    Returns
    -------

    dict of all input parameters of the MVS E-Lands simulation
    """

    with open(path_input_file) as json_file:
        dict_values = json.load(json_file)

    # Retrieve the simulation setting in the right format
    if SIMULATION_SETTINGS in dict_values:
        dict_values[SIMULATION_SETTINGS] = convert_from_json_to_special_types(
            dict_values[SIMULATION_SETTINGS]
        )
        # Compute the END_DATE and the TIME_INDEX
        retrieve_date_time_info(dict_values[SIMULATION_SETTINGS])

        time_index = dict_values[SIMULATION_SETTINGS][TIME_INDEX]
    else:
        time_index = None

    # Convert the values inside the dict to python types
    dict_values = convert_from_json_to_special_types(dict_values, time_index=time_index)

    # The user specified a value
    if path_input_folder is not None:
        dict_values[SIMULATION_SETTINGS][PATH_INPUT_FOLDER] = path_input_folder

    # The user specified a value
    if path_output_folder is not None:
        dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER] = path_output_folder
        dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER_INPUTS] = os.path.join(
            path_output_folder, INPUTS_COPY
        )

    # Move the json file created from csv to the copy of the input folder in the output folder
    if move_copy is True:
        os.replace(
            path_input_file,
            os.path.join(
                dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER_INPUTS], CSV_FNAME,
            ),
        )

    # add default value if the field PATHS_TO_PLOTS is not already present
    if PATHS_TO_PLOTS not in dict_values:
        dict_values.update(copy.deepcopy(DICT_PLOTS))

    # raise a warning if required parameters are missing, see REQUIRED_MVS_PARAMETERS in
    # constants.py for more information, note json and csv required parameter are independent from
    # one another at the moment
    compare_input_parameters_with_reference(
        dict_values, flag_missing=flag_missing_values, set_default=set_default_values
    )

    return dict_values



"""
Helper functions
================

Util functions that are useful throughout the MVS

Including:
- find_valvue_by_key(): Finds value of a key in a nested dictionary.
"""

import os


def find_value_by_key(data, target, result=None):
    """
    Finds value of a key in a nested dictionary.

    Parameters
    ----------
    data: dict
        Dict to be searched for target key

    target: str
        Key for which the value should be found in data

    result: None, value or list
        Only provided if function loops in itself

    Returns
    -------
    value if the key is only once in data
    list of values if it appears multiple times.
    """
    # check each item-value pair in the level
    for k, v in data.items():
        # if target in keys of level
        if k == target:
            if result is None:
                result = v
            elif isinstance(result, list):
                # Expands list of key finds
                result.append(v)
            else:
                # creates list for multiple key finds
                previous_result = result
                result = []
                result.append(previous_result)
                result.append(v)
        # Check next level for target
        if isinstance(v, dict):
            result = find_value_by_key(data=v, target=target, result=result)
    return result


def translates_epa_strings_to_mvs_readable(folder_name, file_name):
    """
    This function translates the json file generated by the EPA to a file readable by the MVS.
    This is necessary as there are some parameter names whose string representative differs in both tools.

    Parameters
    ----------
    folder_name: path
        Path to the folder with the json file
    file_name: json file name with extension
        Json to be converted

    Returns
    -------
    Stores converted json file to current dict

    Usage:
        `import multi_vector_simulator.utils.helpers as helpers`
        `helpers.translates_epa_strings_to_mvs_readable("./epa_benchmark", "epa_benchmark.json-original")`
    """
    import json
    from multi_vector_simulator.utils.data_parser import convert_epa_params_to_mvs

    with open(os.path.join(folder_name, file_name)) as json_file:
        epa_dict = json.load(json_file)

    dict_values = convert_epa_params_to_mvs(epa_dict)

    with open(os.path.join(folder_name, "mvs_config.json"), "w") as json_file:
        json.dump(dict_values, json_file, indent=4)


def get_item_if_list(list_or_float, index):
    if isinstance(list_or_float, list):
        answer = list_or_float[index]
    else:
        answer = list_or_float
    return answer


def get_length_if_list(list_or_float):
    if isinstance(list_or_float, list):
        answer = len(list_or_float)
    else:
        answer = 0
    return answer



"""
Output definition
=================

This file contains multiple lists of paramters that are written to the "scalars.xlsx" output file of the MVS.
One should simply import new KPI names from constants.py or constants_json_strings.py
and add them to the list to make them appear in the excel output sheet "scalars".
"""

from multi_vector_simulator.utils.constants import (
    COST_TOTAL,
    COST_OPERATIONAL_TOTAL,
    COST_INVESTMENT,
    COST_UPFRONT,
    COST_REPLACEMENT,
    COST_DISPATCH,
    COST_OM,
    ANNUITY_TOTAL,
    ANNUITY_OM,
    UNIT,
    LCOE_ASSET,
    INSTALLED_CAP,
    OPTIMIZED_ADD_CAP,
    TOTAL_FLOW,
    ANNUAL_TOTAL_FLOW,
    PEAK_FLOW,
    AVERAGE_FLOW,
    TOTAL_EMISSIONS,
)

from multi_vector_simulator.utils.constants_json_strings import LABEL

######################
# Tab "cost_matrix"  #
######################
KPI_COST_MATRIX_ENTRIES = [
    LABEL,
    COST_TOTAL,
    COST_OPERATIONAL_TOTAL,
    COST_INVESTMENT,
    COST_UPFRONT,
    COST_REPLACEMENT,
    COST_DISPATCH,
    COST_OM,
    ANNUITY_TOTAL,
    ANNUITY_OM,
    LCOE_ASSET,
]

########################
# Tab "scalar_matrix"  #
########################

KPI_SCALAR_MATRIX_ENTRIES = [
    LABEL,
    UNIT,
    INSTALLED_CAP,
    OPTIMIZED_ADD_CAP,
    TOTAL_FLOW,
    ANNUAL_TOTAL_FLOW,
    PEAK_FLOW,
    AVERAGE_FLOW,
    TOTAL_EMISSIONS,
]
##################
# Tab "scalars"  #
##################

#############################
# KPI "individual_sectors"  #
#############################



"""
Module D0 - Model building
==========================

Functional requirements of module D0:
- measure time needed to build model
- measure time needed to solve model
- generate energy system model for oemof
- create dictionary of components so that they can be used for constraints and some
- raise warning if component not a (in mvs defined) oemof model type
- add all energy conversion, energy consumption, energy production, energy storage devices model
- plot network graph
- at constraints to remote model
- store lp file (optional)
- start oemof simulation
- process results by giving them to the next function
- dump oemof results
- add simulation parameters to dict values
"""

import logging
import os
import timeit
import warnings

from oemof.solph import processing, network
import oemof.solph as solph

import multi_vector_simulator.D1_model_components as D1
import multi_vector_simulator.D2_model_constraints as D2

from multi_vector_simulator.utils.constants import (
    PATH_OUTPUT_FOLDER,
    ES_GRAPH,
    PATHS_TO_PLOTS,
    PLOT_SANKEY,
    PLOTS_ES,
    LP_FILE,
)
from multi_vector_simulator.utils.constants_json_strings import (
    ENERGY_BUSSES,
    ENERGY_VECTOR,
    OEMOF_ASSET_TYPE,
    ACCEPTED_ASSETS_FOR_ASSET_GROUPS,
    OEMOF_GEN_STORAGE,
    OEMOF_SINK,
    OEMOF_SOURCE,
    OEMOF_TRANSFORMER,
    OEMOF_BUSSES,
    OEMOF_ExtractionTurbineCHP,
    VALUE,
    SIMULATION_SETTINGS,
    LABEL,
    TIME_INDEX,
    OUTPUT_LP_FILE,
    SIMULATION_RESULTS,
    OBJECTIVE_VALUE,
    SIMULTATION_TIME,
    MODELLING_TIME,
)

from multi_vector_simulator.utils.exceptions import (
    MVSOemofError,
    WrongOemofAssetForGroupError,
    UnknownOemofAssetType,
)


def run_oemof(dict_values, save_energy_system_graph=False, return_les=False):
    """
    Creates and solves energy system model generated from excel template inputs.
    Each component is included by calling its constructor function in D1_model_components.

    Parameters
    ----------
    dict values: dict
        Includes all dictionary values describing the whole project, including costs,
        technical parameters and components. In C0_data_processing, each component was attributed
        with a certain in/output bus.

    save_energy_system_graph: bool
        if set to True, saves a local copy of the energy system's graph

    return_les: bool
        if set to True, the return also includes the local_energy_system in third position

    Returns
    -------
    saves and returns oemof simulation results
    """

    start = timer.initalize()

    model, dict_model = model_building.initialize(dict_values)

    model = model_building.adding_assets_to_energysystem_model(
        dict_values, dict_model, model
    )

    model_building.plot_networkx_graph(
        dict_values, model, save_energy_system_graph=save_energy_system_graph
    )

    logging.debug("Creating oemof model based on created components and busses...")
    local_energy_system = solph.Model(model)
    logging.debug("Created oemof model based on created components and busses.")

    local_energy_system = D2.add_constraints(
        local_energy_system, dict_values, dict_model
    )
    model_building.store_lp_file(dict_values, local_energy_system)

    model, results_main, results_meta = model_building.simulating(
        dict_values, model, local_energy_system
    )

    model_building.plot_sankey_diagramm(
        dict_values, model, save_energy_system_graph=save_energy_system_graph
    )

    timer.stop(dict_values, start)

    if return_les is True:
        return results_meta, results_main, local_energy_system
    else:
        return results_meta, results_main


class model_building:
    def initialize(dict_values):
        """
        Initalization of oemof model

        Parameters
        ----------
        dict_values: dict
            dictionary of simulation

        Returns
        -------
        oemof energy model (oemof.solph.network.EnergySystem), dict_model which gathers the assets added to this model later.
        """
        logging.info("Initializing oemof simulation.")
        model = solph.EnergySystem(
            timeindex=dict_values[SIMULATION_SETTINGS][TIME_INDEX]
        )

        # this dictionary will include all generated oemof objects
        dict_model = {
            OEMOF_BUSSES: {},
            OEMOF_SINK: {},
            OEMOF_SOURCE: {},
            OEMOF_TRANSFORMER: {},
            OEMOF_GEN_STORAGE: {},
            OEMOF_ExtractionTurbineCHP: {},
        }

        return model, dict_model

    def adding_assets_to_energysystem_model(dict_values, dict_model, model, **kwargs):
        """

        Parameters
        ----------
        dict_values: dict
            dict of simulation data

        dict_model:
            Updated list of assets in the oemof energy system model

        model: oemof.solph.network.EnergySystem
            Model of oemof energy system

        Returns
        -------

        """
        logging.info("Adding components to oemof energy system model...")

        # Busses have to be defined first
        for bus in dict_values[ENERGY_BUSSES]:
            D1.bus(
                model,
                dict_values[ENERGY_BUSSES][bus][LABEL],
                energy_vector=dict_values[ENERGY_BUSSES][bus][ENERGY_VECTOR],
                **dict_model,
            )

        # Adding step by step all assets defined within the asset groups
        for asset_group in ACCEPTED_ASSETS_FOR_ASSET_GROUPS:
            if asset_group in dict_values:
                for asset in dict_values[asset_group]:
                    type = dict_values[asset_group][asset][OEMOF_ASSET_TYPE]
                    # Checking if the asset type is one accepted for the asset group (security measure)
                    if type in ACCEPTED_ASSETS_FOR_ASSET_GROUPS[asset_group]:
                        # if so, then the appropriate function of D1 should be called
                        if type == OEMOF_TRANSFORMER:
                            D1.transformer(
                                model, dict_values[asset_group][asset], **dict_model
                            )
                        elif type == OEMOF_ExtractionTurbineCHP:
                            D1.chp(model, dict_values[asset_group][asset], **dict_model)
                        elif type == OEMOF_SINK:
                            D1.sink(
                                model, dict_values[asset_group][asset], **dict_model
                            )
                        elif type == OEMOF_SOURCE:
                            D1.source(
                                model, dict_values[asset_group][asset], **dict_model
                            )
                        elif type == OEMOF_GEN_STORAGE:
                            D1.storage(
                                model, dict_values[asset_group][asset], **dict_model
                            )
                        else:
                            raise UnknownOemofAssetType(
                                f"Asset {asset} has type {type}, "
                                f"but this type is not a defined oemof asset type."
                            )

                    else:
                        raise WrongOemofAssetForGroupError(
                            f"Asset {asset} has type {type}, "
                            f"but this type is not an asset type attributed to asset group {asset_group}"
                            f" for oemof model generation."
                        )

        logging.debug("All components added.")
        return model

    def plot_networkx_graph(dict_values, model, save_energy_system_graph=False):
        """
        Plots a graph of the energy system if that graph is to be displayed or stored.

        Parameters
        ----------
        dict_values: dict
            All simulation inputs

        model: `oemof.solph.network.EnergySystem`
            oemof-solph object for energy system model

        save_energy_system_graph: bool
            if True, save the graph in the mvs output folder
            Default: False

        Returns
        -------
        None
        """
        if save_energy_system_graph is True:
            from multi_vector_simulator.F1_plotting import ESGraphRenderer

            fpath = os.path.join(
                dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER], ES_GRAPH
            )
            dict_values[PATHS_TO_PLOTS][PLOTS_ES] += str(fpath)

            # Draw the energy system model
            graph = ESGraphRenderer(model, filepath=fpath)
            logging.debug("Created graph of the energy system model.")

            graph.render()

    def plot_sankey_diagramm(dict_values, model, save_energy_system_graph=False):
        """
        Prepare a sankey diagram of the simulated energy model

        Parameters
        ----------
        dict_values: dict
            All simulation inputs

         model: `oemof.solph.network.EnergySystem`
            oemof-solph object for energy system model

        save_energy_system_graph: bool
            if True, save the graph in the mvs output folder
            Default: False

        Returns
        -------

        """
        if save_energy_system_graph is True:
            from multi_vector_simulator.F1_plotting import ESGraphRenderer

            graph = ESGraphRenderer(model)
            dict_values[PATHS_TO_PLOTS][PLOT_SANKEY] = graph.sankey(
                model.results["main"]
            )

    def store_lp_file(dict_values, local_energy_system):
        """
        Stores linear equation system generated with pyomo as an "lp file".

        Parameters
        ----------
        dict_values: dict
            All simulation input data

        local_energy_system: object
            pyomo object including all constraints of the energy system

        Returns
        -------
        Nothing.
        """
        if dict_values[SIMULATION_SETTINGS][OUTPUT_LP_FILE][VALUE] is True:
            path_lp_file = os.path.join(
                dict_values[SIMULATION_SETTINGS][PATH_OUTPUT_FOLDER], LP_FILE
            )
            logging.debug("Saving to lp-file.")
            local_energy_system.write(
                path_lp_file, io_options={"symbolic_solver_labels": True},
            )

    def simulating(dict_values, model, local_energy_system):
        """
        Initiates the oemof-solph simulation, accesses results and writes main results into dict

        If an error is encountered in the oemof solver, mvs should not be allowed to continue,
        otherwise other errors related to the uncomplete simulation result might occur and it will
        be more obscure to the endusers what went wrong.

        A MVS error is raised if the omoef solver warning states explicitely that
        "termination condition infeasible", otherwise the oemof solver warning is re-raised as
        an error.


        Parameters
        ----------
        dict_values: dict
            All simulation inputs

        model: object
            oemof-solph object for energy system model

        local_energy_system: object
            pyomo object storing all constraints of the energy system model

        Returns
        -------
        Updated model with results, main results (flows, assets) and meta results (simulation)
        """

        logging.info("Starting simulation.")
        # turn warnings into errors
        warnings.filterwarnings("error")
        try:
            local_energy_system.solve(
                solver="cbc",
                solve_kwargs={
                    "tee": False
                },  # if tee_switch is true solver messages will be displayed
                cmdline_options={"ratioGap": str(0.03)},
            )  # ratioGap allowedGap mipgap
        except UserWarning as e:
            error_message = str(e)
            compare_message = "termination condition infeasible"
            if compare_message in error_message:
                error_message = (
                    f"The following error occurred during the mvs solver: {error_message}\n\n "
                    f"There are several reasons why this could have happened."
                    "\n\t- the energy system is not properly connected. "
                    "\n\t- the capacity of some assets might not have been optimized. "
                    "\n\t- the demands might not be supplied with the installed capacities in "
                    "current energy system. Check your maximum power demand and if your energy "
                    "production assets and/or energy conversion assets have enough capacity to "
                    "meet the total demand"
                )
                logging.error(error_message)
                raise MVSOemofError(error_message) from None
            else:
                raise e
        # stop turning warnings into errors
        warnings.resetwarnings()

        # add results to the energy system to make it possible to store them.
        results_main = processing.results(local_energy_system)
        results_meta = processing.meta_results(local_energy_system)

        model.results["main"] = results_main
        model.results["meta"] = results_meta

        dict_values.update(
            {
                SIMULATION_RESULTS: {
                    LABEL: SIMULATION_RESULTS,
                    OBJECTIVE_VALUE: results_meta["objective"],
                    SIMULTATION_TIME: round(results_meta["solver"]["Time"], 2),
                }
            }
        )
        logging.info(
            "Simulation time: %s minutes.",
            round(dict_values[SIMULATION_RESULTS][SIMULTATION_TIME] / 60, 2),
        )
        return model, results_main, results_main


class timer:
    def initalize():
        """
        Starts a timer
        Returns
        -------
        """
        # Start clock to determine total simulation time
        start = timeit.default_timer()
        return start

    def stop(dict_values, start):
        """
        Ends timer and adds duration of simulation to dict_values
        Parameters
        ----------
        dict_values: dict
            Dict of simulation including SIMULATION_RESULTS key
        start: timestamp
            start time of timer

        Returns
        -------
        Simulation time in dict_values
        """
        duration = timeit.default_timer() - start
        dict_values[SIMULATION_RESULTS].update({MODELLING_TIME: round(duration, 2)})

        logging.info(
            "Modeling time: %s minutes.",
            round(dict_values[SIMULATION_RESULTS][MODELLING_TIME] / 60, 2),
        )



"""
This is the main file of the tool "Multi-vector simulation tool".

Tool structure:

(parent)    multi_vector_simulator.py

(child)    constants.py

(child)     --A0_initialization.py

(child)      --B0_data_input.py

(child)     --C0_data_processing.py
(child sub)    --C1_verification.py
(child sub)    --C2_economic_processing.py

(child)     --D0_modelling_and_optimization.py
(child sub)    --D1_model_components.py
(child sub)    --D2_model_constraints.py

(child)     --E0_evaluation.py
(child sub)    --E1_process_results.py
(child sub)    --E2_verification_of_constraints.py
(child sub)    --E4_verification.py

(child)    --F0_output.py
(child sub)    --F1_plotting.py
(child sub)    --F2_autoreport.py
patent:     Main file, all children connected through parent
child:      Child file, one of the main functions of the tool.
            Internal processes, feeds output back to parent
child-sub:  Sub-child function, feeds only back to child functions
"""

import logging
import json
import os
import tempfile

from oemof.tools import logger

# Loading all child functions
import multi_vector_simulator.B0_data_input_json as B0
import multi_vector_simulator.C0_data_processing as C0
import multi_vector_simulator.D0_modelling_and_optimization as D0
import multi_vector_simulator.E0_evaluation as E0
import multi_vector_simulator.F0_output as F0
from multi_vector_simulator.version import version_num, version_date
from multi_vector_simulator.utils import (
    data_parser,
    nested_dict_crawler,
    get_nested_value,
)


from multi_vector_simulator.utils.constants_json_strings import (
    SIMULATION_SETTINGS,
    OUTPUT_LP_FILE,
    VALUE,
    UNIT,
)
from multi_vector_simulator.utils.constants import TYPE_STR


def run_simulation(json_dict, epa_format=True, **kwargs):
    r"""
     Starts MVS tool simulation from an input json file

     Parameters
    -----------
     json_dict: dict
         json from http request
     epa_format: bool, optional
         Specifies whether the output is formatted for EPA standards
         Default: True

     Other Parameters
     ----------------
     pdf_report: bool, optional
         Can generate an automatic pdf report of the simulation's results (True) or not (False)
         Default: False.
     display_output : str, optional
         Sets the level of displayed logging messages.
         Options: "debug", "info", "warning", "error". Default: "info".
     lp_file_output : bool, optional
         Specifies whether linear equation system generated is saved as lp file.
         Default: False.

    """
    display_output = kwargs.get("display_output", None)

    if display_output == "debug":
        screen_level = logging.DEBUG
    elif display_output == "info":
        screen_level = logging.INFO
    elif display_output == "warning":
        screen_level = logging.WARNING
    elif display_output == "error":
        screen_level = logging.ERROR
    else:
        screen_level = logging.INFO

    # Define logging settings and path for saving log
    logger.define_logging(screen_level=screen_level)

    welcome_text = (
        "\n \n Multi-Vector Simulation Tool (MVS) V"
        + version_num
        + " "
        + "\n Version: "
        + version_date
        + " "
        + '\n Part of the toolbox of H2020 project "E-LAND", '
        + "Integrated multi-vector management system for Energy isLANDs"
        + "\n Coded at: Reiner Lemoine Institute (Berlin) "
        + "\n Reference: https://zenodo.org/record/4610237 \n \n "
    )

    logging.info(welcome_text)

    logging.debug("Accessing script: B0_data_input_json")
    dict_values = B0.convert_from_json_to_special_types(json_dict)

    # if True will return the lp file's content in dict_values
    lp_file_output = dict_values[SIMULATION_SETTINGS][OUTPUT_LP_FILE][VALUE]
    # to avoid the lp file being saved somewhere on the server
    dict_values[SIMULATION_SETTINGS][OUTPUT_LP_FILE][VALUE] = False

    print("")
    logging.debug("Accessing script: C0_data_processing")
    C0.all(dict_values)

    print("")
    logging.debug("Accessing script: D0_modelling_and_optimization")
    results_meta, results_main, local_energy_system = D0.run_oemof(
        dict_values, return_les=True
    )

    if lp_file_output is True:
        logging.debug("Saving the content of the model's lp file")
        with tempfile.TemporaryDirectory() as tmpdirname:
            local_energy_system.write(
                os.path.join(tmpdirname, "lp_file.lp"),
                io_options={"symbolic_solver_labels": True},
            )
            with open(os.path.join(tmpdirname, "lp_file.lp")) as fp:
                file_content = fp.read()

        dict_values[SIMULATION_SETTINGS][OUTPUT_LP_FILE][VALUE] = file_content
        dict_values[SIMULATION_SETTINGS][OUTPUT_LP_FILE][UNIT] = TYPE_STR

    print("")
    logging.debug("Accessing script: E0_evaluation")
    E0.evaluate_dict(dict_values, results_main, results_meta)

    logging.debug("Convert results to json")

    if epa_format is True:
        epa_dict_values = data_parser.convert_mvs_params_to_epa(dict_values)

        json_values = F0.store_as_json(epa_dict_values)
        answer = json.loads(json_values)

    else:
        answer = dict_values

    return answer


def run_sensitivity_analysis_step(
    json_input, step_idx, output_variables, epa_format=True, **kwargs
):
    r"""
     Starts MVS tool simulation from an input json file

     Parameters
    -----------
    json_input: dict
        json from http request
    step_idx: int
        step of the sensitivity analysis
    output_variables: tuple of str
        collection of output variables names
    epa_format: bool, optional
        Specifies whether the output is formatted for EPA standards
        Default: True

     Other Parameters
     ----------------
     pdf_report: bool, optional
         Can generate an automatic pdf report of the simulation's results (True) or not (False)
         Default: False.
     display_output : str, optional
         Sets the level of displayed logging messages.
         Options: "debug", "info", "warning", "error". Default: "info".
     lp_file_output : bool, optional
         Specifies whether linear equation system generated is saved as lp file.
         Default: False.

    """

    # Process the argument json_input based on its type
    if isinstance(json_input, str):
        # load the file if it is a path
        simulation_input = B0.load_json(json_input)
    elif isinstance(json_input, dict):
        # this is already a json variable
        simulation_input = json_input
    else:
        simulation_input = None
        logging.error(
            f"Simulation input `{json_input}` is neither a file path, nor a json dict. "
            f"It can therefore not be processed."
        )

    sim_output_json = run_simulation(
        simulation_input, display_output="error", epa_format=epa_format
    )
    output_variables_paths = nested_dict_crawler(sim_output_json)

    output_parameters = {}
    # for each of the output parameter path, add the value located under this path in
    # the final json dict, that could also be applied to the full json dict as
    # post-processing
    for output_param in output_variables:
        output_param_pathes = output_variables_paths.get(output_param, [])

        if len(output_param_pathes) == 0:
            output_parameters[output_param] = dict(
                value=None,
                path=f"Not found in mvs results (version {version_num}), check if you have typos in output parameter name",
            )

        for output_param_path in output_param_pathes:
            if output_param not in output_parameters:
                output_parameters[output_param] = dict(
                    value=[get_nested_value(sim_output_json, output_param_path)],
                    path=[".".join(output_param_path)],
                )
            else:
                output_parameters[output_param]["value"].append(
                    get_nested_value(sim_output_json, output_param_path)
                )
                output_parameters[output_param]["path"].append(
                    ".".join(output_param_path)
                )

    return {"step_idx": step_idx, "output_values": output_parameters}



import json
import logging
from multi_vector_simulator.utils import (
    get_nested_value,
    set_nested_value,
    split_nested_path,
)
from multi_vector_simulator.server import run_simulation
from multi_vector_simulator.B0_data_input_json import (
    load_json,
    convert_from_json_to_special_types,
)


def single_param_variation_analysis(
    param_values, json_input, json_path_to_param_value, json_path_to_output_value=None
):
    r"""Run mvs simulations by varying one of the input parameters to access output's sensitivity

    Parameters
    ----------
    param_values: list of values (type can vary)
    json_input: path or dict
        input parameters for the multi-vector simulation
    json_path_to_param_value: tuple or str
        succession of keys which lead the value of the parameter to vary in the json_input dict
        potentially nested structure. The order of keys is to be read from left to right. In the
        case of str, each key should be separated by a `.` or a `,`.
    json_path_to_output_value: tuple of tuple or str, optional
        collection of succession of keys which lead the value of an output parameter of interest in
        the json dict of the simulation's output. The order of keys is to be read from left to
        right. In the case of str, each key should be separated by a `.` or a `,`.

    Returns
    -------
    The simulation output json matched to the list of variied parameter values
    """

    # Process the argument json_input based on its type
    if isinstance(json_input, str):
        # load the file if it is a path
        simulation_input = load_json(json_input)
    elif isinstance(json_input, dict):
        # this is already a json variable
        simulation_input = json_input
    else:
        simulation_input = None
        logging.error(
            f"Simulation input `{json_input}` is neither a file path, nor a json dict. "
            f"It can therefore not be processed."
        )
    param_path_tuple = split_nested_path(json_path_to_param_value)
    answer = []
    if simulation_input is not None:
        for param_val in param_values:
            # modify the value of the parameter before running a new simulation
            modified_input = set_nested_value(
                simulation_input, param_val, param_path_tuple
            )
            # run a simulation with next value of the variable parameter and convert the result to
            # mvs special json type
            sim_output_json = run_simulation(
                modified_input, display_output="error", epa_format=False
            )
            print(sim_output_json)
            if json_path_to_output_value is None:
                answer.append(sim_output_json)
            else:
                output_parameters = {}
                # for each of the output parameter path, add the value located under this path in
                # the final json dict, that could also be applied to the full json dict as
                # post-processing
                for output_param in json_path_to_output_value:
                    output_param = split_nested_path(output_param)
                    output_parameters[output_param] = get_nested_value(
                        sim_output_json, output_param
                    )
                answer.append(output_parameters)

    return {"parameters": param_values, "outputs": answer}
